import xtrack as xt
import numpy as np

assert_allclose = np.testing.assert_allclose

slice_mode = 'thin'
tilted = True
orientation = 'acw'
transform_to_actual_elements = True

line = xt.Line(
    elements=[
        xt.Drift(),
        xt.Bend(),
        xt.Drift(),
        xt.Bend(),
        xt.Drift(),
        xt.Bend(),
        xt.Drift(),
        xt.Bend(),
    ]
)
line.build_tracker()

line.vars['l_drift'] = 999.
line.vars['l_bend'] = 999.
line.vars['h_bend'] = 999.
line.vars['tilt_bend_deg'] = 999.


line.element_refs['e0'].length = line.vars['l_drift']
line.element_refs['e1'].length = line.vars['l_bend']
line.element_refs['e2'].length = line.vars['l_drift']
line.element_refs['e3'].length = line.vars['l_bend']
line.element_refs['e4'].length = line.vars['l_drift']
line.element_refs['e5'].length = line.vars['l_bend']
line.element_refs['e6'].length = line.vars['l_drift']
line.element_refs['e7'].length = line.vars['l_bend']

line.element_refs['e1'].h = line.vars['h_bend']
line.element_refs['e3'].h = line.vars['h_bend']
line.element_refs['e5'].h = line.vars['h_bend']
line.element_refs['e7'].h = line.vars['h_bend']

line.element_refs['e1'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
line.element_refs['e3'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
line.element_refs['e5'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180
line.element_refs['e7'].rot_s_rad = line.vars['tilt_bend_deg'] * np.pi / 180

line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Teapot(20, mode=slice_mode))])
line.build_tracker()

line.vars['l_drift'] = 1
line.vars['l_bend'] = 1
if orientation == 'cw':
    line.vars['h_bend'] = np.pi/2 / line.vars['l_bend']
elif orientation == 'acw':
    line.vars['h_bend'] = -np.pi/2 / line.vars['l_bend']

if tilted:
    line.vars['tilt_bend_deg'] = 90
else:
    line.vars['tilt_bend_deg'] = 0

if slice_mode == 'thin' and transform_to_actual_elements:
    line.discard_tracker()
    line._replace_with_equivalent_elements()
    line.build_tracker()
    assert isinstance(line['e1..1'], xt.Multipole)

sv = line.survey()
assert_allclose(sv.Z[-1], 0, rtol=0, atol=1e-13)
assert_allclose(sv.X[-1], 0, rtol=0, atol=1e-13)
assert_allclose(sv.Y[-1], 0, rtol=0, atol=1e-13)
assert_allclose(sv.s[-1], 8, rtol=0, atol=1e-13)


if not tilted and orientation == 'acw':
    assert_allclose(np.abs(sv.Y), 0, rtol=0, atol=1e-14)
    assert_allclose(np.trapz(sv.X, sv.Z), -4.818 , rtol=0, atol=2e-3) # clockwise
elif not tilted and orientation == 'cw':
    assert_allclose(np.abs(sv.Y), 0, rtol=0, atol=1e-14)
    assert_allclose(np.trapz(sv.X, sv.Z), 4.818 , rtol=0, atol=2e-3)
elif tilted and orientation == 'acw':
    assert_allclose(np.abs(sv.X), 0, rtol=0, atol=1e-14)
    assert_allclose(np.trapz(sv.Y, sv.Z), -4.818 , rtol=0, atol=2e-3)
elif tilted and orientation == 'cw':
    assert_allclose(np.abs(sv.X), 0, rtol=0, atol=1e-14)
    assert_allclose(np.trapz(sv.Y, sv.Z), 4.818 , rtol=0, atol=2e-3)


import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(sv.Z, sv.X)
plt.plot(sv.Z, sv.Y)

plt.figure(2)
plt.plot(sv.s, sv.X)
plt.plot(sv.s, sv.Y)
plt.plot(sv.s, sv.Z)

plt.show()


import xtrack as xt
import numpy as np

line = xt.Line.from_json("../../test_data/hllhc15_thick/lhc_thick_with_knobs.json")

opt = line.match(
    start='mq.30l8.b1', end='mq.23l8.b1',
    betx=1, bety=1, y=0, py=0,
    vary=[xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1',],
                    step=1e-10, limits=[-1e-3, 1e-3]),
          xt.VaryList(['acbv26.l8b1', 'acbv24.l8b1'],
                    step=1e-10, limits=[-1e-3, 1e-3], tag='mytag')],
    targets = [
        xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
        xt.TargetSet(y=0, py=0, at=xt.END)
    ])

# Check autogenerated tags
assert opt.targets[0].tag == 'mb.b28l8.b1_y'
assert opt.targets[1].tag == 'mb.b28l8.b1_py'
assert opt.targets[2].tag == 'END_y'
assert opt.targets[3].tag == 'END_py'

# Check target getitem
assert opt.targets[0] == opt.targets['mb.b28l8.b1_y']
assert opt.targets[1] == opt.targets['mb.b28l8.b1_py']
assert opt.targets[2] == opt.targets['END_y']
assert opt.targets[3] == opt.targets['END_py']

assert opt.targets['END.*'][0] is opt.targets[2]
assert opt.targets['END.*'][1] is opt.targets[3]

assert np.all(
    opt.targets.status(ret=True).tag == opt.target_status(ret=True).tag)

assert opt.vary['myt.*'][0] is opt.vary[2]
assert opt.vary['myt.*'][1] is opt.vary[3]
assert np.all(opt.vary.status(ret=True).tag == np.array(['', '', 'mytag', 'mytag']))
assert np.all(opt.vary.status(ret=True).tag == opt.vary_status(ret=True).tag)
assert np.all(opt.vary.status(ret=True).name == np.array(
    ['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1']))

opt1 = opt.clone(name='opt1')
assert opt1.name == 'opt1'
assert str(opt1.targets[0]) == str(opt.targets[0])
assert str(opt1.targets[1]) == str(opt.targets[1])
assert str(opt1.targets[2]) == str(opt.targets[2])
assert str(opt1.targets[3]) == str(opt.targets[3])
assert str(opt1.vary[0]) == str(opt.vary[0])
assert str(opt1.vary[1]) == str(opt.vary[1])
assert str(opt1.vary[2]) == str(opt.vary[2])
assert str(opt1.vary[3]) == str(opt.vary[3])

opt2 = opt.clone(name='opt2', remove_vary=True)
assert len(opt2.vary) == 0
assert str(opt2.targets[0]) == str(opt.targets[0])
assert str(opt2.targets[1]) == str(opt.targets[1])
assert str(opt2.targets[2]) == str(opt.targets[2])
assert str(opt2.targets[3]) == str(opt.targets[3])

opt3 = opt.clone(name='opt3', remove_targets=True,
                 add_targets=[xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1')])
assert len(opt3.targets) == 2
assert opt.targets[0].tag == 'mb.b28l8.b1_y'
assert opt.targets[1].tag == 'mb.b28l8.b1_py'
assert str(opt3.vary[0]) == str(opt.vary[0])
assert str(opt3.vary[1]) == str(opt.vary[1])
assert str(opt3.vary[2]) == str(opt.vary[2])
assert str(opt3.vary[3]) == str(opt.vary[3])

opt4 = opt.clone(name='opt4', remove_vary=True, remove_targets=False,
                    add_vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1']),
                    )
assert len(opt4.vary) == 2
assert opt.vary[0].name == 'acbv30.l8b1'
assert opt.vary[1].name == 'acbv28.l8b1'


import time
import numpy as np

import xtrack as xt
from xtrack.slicing import Teapot, Strategy

line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.twiss_default['method'] = '4d'
line.build_tracker()

line_thick = line.copy()
line_thick.build_tracker()

slicing_strategies = [
    Strategy(slicing=Teapot(1)),  # Default
    Strategy(slicing=Teapot(4), element_type=xt.Bend),
    Strategy(slicing=Teapot(20), element_type=xt.Quadrupole),
    Strategy(slicing=Teapot(2), name=r'^mb\..*'),
    Strategy(slicing=Teapot(5), name=r'^mq\..*'),
    Strategy(slicing=Teapot(2), name=r'^mqt.*'),
    Strategy(slicing=Teapot(60), name=r'^mqx.*'),
]

line.discard_tracker()
t1 = time.time()
line.slice_thick_elements(slicing_strategies=slicing_strategies)
t2 = time.time()
print('\nTime slice thick: ', t2-t1)
line.build_tracker()

tw = line.twiss()
tw_thick = line_thick.twiss()

beta_beat_x_at_ips = [tw['betx', f'ip{nn}'] / tw_thick['betx', f'ip{nn}'] - 1
                        for nn in range(1, 9)]
beta_beat_y_at_ips = [tw['bety', f'ip{nn}'] / tw_thick['bety', f'ip{nn}'] - 1
                        for nn in range(1, 9)]

assert np.allclose(beta_beat_x_at_ips, 0, atol=3e-3)
assert np.allclose(beta_beat_y_at_ips, 0, atol=3e-3)

# Checks on orbit knobs
assert np.isclose(tw_thick['px', 'ip1'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw_thick['py', 'ip1'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw_thick['px', 'ip5'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw_thick['py', 'ip5'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw['px', 'ip1'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw['py', 'ip1'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw['px', 'ip5'], 0, rtol=0, atol=1e-7)
assert np.isclose(tw['py', 'ip5'], 0, rtol=0, atol=1e-7)

line.vars['on_x1'] = 50
line.vars['on_x5'] = 60
line_thick.vars['on_x1'] = 50
line_thick.vars['on_x5'] = 60

tw = line.twiss()
tw_thick = line_thick.twiss()

assert np.isclose(tw_thick['px', 'ip1'], 50e-6, rtol=0, atol=5e-7)
assert np.isclose(tw_thick['py', 'ip1'], 0, rtol=0, atol=5e-7)
assert np.isclose(tw_thick['px', 'ip5'], 0, rtol=0, atol=5e-7)
assert np.isclose(tw_thick['py', 'ip5'], 60e-6, rtol=0, atol=5e-7)
assert np.isclose(tw['px', 'ip1'], 50e-6, rtol=0, atol=5e-7)
assert np.isclose(tw['py', 'ip1'], 0, rtol=0, atol=5e-7)
assert np.isclose(tw['px', 'ip5'], 0, rtol=0, atol=5e-7)
assert np.isclose(tw['py', 'ip5'], 60e-6, rtol=0, atol=5e-7)

t1 = time.time()
opt_thick = line_thick.match(
    vary=[
        xt.Vary('kqtf.b1', step=1e-8),
        xt.Vary('kqtd.b1', step=1e-8),
        xt.Vary('ksf.b1', step=1e-8),
        xt.Vary('ksd.b1', step=1e-8),
    ],
    targets = [
        xt.Target('qx', 62.27, tol=1e-4),
        xt.Target('qy', 60.29, tol=1e-4),
        xt.Target('dqx', 10.0, tol=0.05),
        xt.Target('dqy', 12.0, tol=0.05)])
t2 = time.time()
print('\nTime match thick: ', t2-t1)

opt_thin = line.match(
    vary=[
        xt.Vary('kqtf.b1', step=1e-8),
        xt.Vary('kqtd.b1', step=1e-8),
        xt.Vary('ksf.b1', step=1e-8),
        xt.Vary('ksd.b1', step=1e-8),
    ],
    targets = [
        xt.Target('qx', 62.27, tol=1e-4),
        xt.Target('qy', 60.29, tol=1e-4),
        xt.Target('dqx', 10.0, tol=0.05),
        xt.Target('dqy', 12.0, tol=0.05)])
t2 = time.time()
print('\nTime match thin: ', t2-t1)

line.to_json('lhc_thin.json')

tw = line.twiss()
tw_thick = line_thick.twiss()

assert np.isclose(tw_thick.qx, 62.27, rtol=0, atol=1e-4)
assert np.isclose(tw_thick.qy, 60.29, rtol=0, atol=1e-4)
assert np.isclose(tw_thick.dqx, 10.0, rtol=0, atol=0.05)
assert np.isclose(tw_thick.dqy, 12.0, rtol=0, atol=0.05)
assert np.isclose(tw.qx, 62.27, rtol=0, atol=1e-4)
assert np.isclose(tw.qy, 60.29, rtol=0, atol=1e-4)
assert np.isclose(tw.dqx, 10.0, rtol=0, atol=0.05)
assert np.isclose(tw.dqy, 12.0, rtol=0, atol=0.05)

assert np.isclose(line.vars['kqtf.b1']._value, line_thick.vars['kqtf.b1']._value,
                    rtol=0.03, atol=0)
assert np.isclose(line.vars['kqtd.b1']._value, line_thick.vars['kqtd.b1']._value,
                    rtol=0.03, atol=0)
assert np.isclose(line.vars['ksf.b1']._value, line_thick.vars['ksf.b1']._value,
                    rtol=0.01, atol=0)
assert np.isclose(line.vars['ksd.b1']._value, line_thick.vars['ksd.b1']._value,
                    rtol=0.01, atol=0)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xobjects as xo

################################################################
# Definition of a beam element with an internal data recording #
################################################################

# We define a data structure to allow all elements of a new BeamElement type
# to store data in one place. Such a data structure needs to contain a field called
# `_index` of type `xtrack.RecordIndex`, which will be used internally to
# keep count of the number of records stored in the data structure. Together with
# the index, the structure can contain an arbitrary number of other fields (which
# need to be arrays) where the data will be stored.

class TestElementRecord(xo.HybridClass):
    _xofields = {
        '_index': xt.RecordIndex,
        'generated_rr': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:]
        }

# The defined data structure can be accessed in the C code of the beam element
# to log data.
# In the following example, the element applies an assigned number of random
# kicks to the horizontal momentum. The internal record is used to store the
# kicks applied together with the corresponding particle_id, turn number and
# element number.

track_method_source = r'''
/*gpufun*/
void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

    // Extract the record and record_index
    TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
    RecordIndex record_index = NULL;
    if (record){
        record_index = TestElementRecordData_getp__index(record);
    }

    int64_t n_kicks = TestElementData_get_n_kicks(el);
    printf("n_kicks %d\n", (int)n_kicks);

    //start_per_particle_block (part0->part)

        for (int64_t i = 0; i < n_kicks; i++) {
            double rr = 1e-6 * RandomUniform_generate(part);
            LocalParticle_add_to_px(part, rr);

            if (record){
                // Get a slot in the record (this is thread safe)
                int64_t i_slot = RecordIndex_get_slot(record_index);
                // The returned slot id is negative if record is NULL or if record is full

                if (i_slot>=0){
                    TestElementRecordData_set_at_element(record, i_slot,
                                                LocalParticle_get_at_element(part));
                    TestElementRecordData_set_at_turn(record, i_slot,
                                                LocalParticle_get_at_turn(part));
                    TestElementRecordData_set_particle_id(record, i_slot,
                                                LocalParticle_get_particle_id(part));
                    TestElementRecordData_set_generated_rr(record, i_slot, rr);
                }
            }
        }


    //end_per_particle_block
}
'''

# To allow elements of a given type to store data in a structure of the type defined
# above we need to add in the element class an attribute called
# `_internal_record_class` to which we bind the data structure type defined above.

class TestElement(xt.BeamElement):
    _xofields={
        'n_kicks': xo.Int64,
        }

    _internal_record_class = TestElementRecord

    _depends_on = [xt.RandomUniform]

    _extra_c_sources = [track_method_source]


# Once these steps are done, the TestElement and its recording feature are ready
# and can be used as follows.

###########################################################
# Usage of a beam element with an internal data recording #
###########################################################

# Line, tracker, particles can be created as usual
line=xt.Line(elements = [
    xt.Drift(length=1.), TestElement(n_kicks=10),
    xt.Drift(length=1.), TestElement(n_kicks=5)])
line.build_tracker()
line._needs_rng = True # Test elements use the random number generator
part = xt.Particles(p0c=6.5e12, x=[1e-3,2e-3,3e-3])

# The record is by default disabled and can be enabled using the following
# dedicated method of the line object. The argument `capacity` defines the
# number of items that can be stored in each element of the internal record
# (tha same space are shared for all the elements of the same type). The returned
# object (called `record` in the following) will be filled with the recorded
# data when the tracker is run. The recording stops when the full capacity is reached.

record = line.start_internal_logging_for_elements_of_type(
                                                    TestElement, capacity=10000)

# Track!
line.track(part, num_turns=10)

# We can now inspect `record`:
#  - `record.generated_rr` contains the random numbers generated by the elements
#  - `record.at_element` contains the element number where the recording took place
#  - `record.at_turn` contains the turn number where the recording took place
#  - `record.particle_id` contains the particle id for which the recording took place
# The number of used slots in `record` can be found in record._index.num_recorded

# The recording can be stopped with the following method:
line.stop_internal_logging_for_elements_of_type(TestElement)

# Track more turns (without logging information in `record`)
line.track(part, num_turns=10)

#!end-doc-part



import xtrack as xt
from xtrack.mad_parser.parse import MadxParser

dct = MadxParser().parse_file('optics.madx')

env = xt.Environment()
env.vars.default_to_zero=True
for vv in dct['vars']:
    env.vars[vv] = dct['vars'][vv]['expr']
env.vars.default_to_zero=False


tt_vars = env.vars.get_table()
out = []
out.append('# Quadrupole strengths:')

out_dct = {}
for nn in tt_vars.rows['kq[^s].*'].name:
    vv = tt_vars['value', nn]
    ee = tt_vars['expr', nn]
    if ee == 'None':
        out_dct[nn] = vv
    else:
        out_dct[nn] = ee

from pprint import pformat
out.append(pformat(out_dct))


with open('optics.py', 'w') as fid:
    fid.write('\n'.join(out))


import xtrack as xt
import numpy as np

k1 = 1.0
k1s = 2.0

length = 0.5

quad = xt.Quadrupole(k1=k1, k1s=k1s, length=length)

n_slices = 1000
ele_thin = []
for ii in range(n_slices):
    ele_thin.append(xt.Drift(length=length/n_slices/2))
    ele_thin.append(xt.Multipole(knl=[0, k1 * length/n_slices],
                                 ksl=[0, k1s * length/n_slices]))
    ele_thin.append(xt.Drift(length=length/n_slices/2))
lref = xt.Line(ele_thin)
lref.build_tracker()

p_test = xt.Particles(gamma0=1.2, x=0.1, y=0.2, delta=0.5)
p_ref = p_test.copy()

quad.track(p_test)
lref.track(p_ref)

assert np.isclose(p_test.x, p_ref.x, atol=1e-8, rtol=0)
assert np.isclose(p_test.px, p_ref.px, atol=5e-8, rtol=0)
assert np.isclose(p_test.y, p_ref.y, atol=1e-8, rtol=0)
assert np.isclose(p_test.py, p_ref.py, atol=5e-8, rtol=0)
assert np.isclose(p_test.zeta, p_ref.zeta, atol=1e-8, rtol=0)
assert np.isclose(p_test.delta, p_ref.delta, atol=5e-8, rtol=0)


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xtrack.particles.pdg as pdg
import xtrack.particles.masses as masses

from xobjects.test_helpers import for_all_test_contexts
import xobjects as xo


def test_names():
    # Test that get_name_from_pdg_id and get_pdg_id_from_name are each other's inverse
    for val in pdg.pdg_table.values():
        name = val[1]
        pdg_id = pdg.get_pdg_id_from_name(name)
        assert pdg.get_name_from_pdg_id(pdg_id, long_name=False) == name
        assert pdg.get_pdg_id_from_name(
            pdg.get_name_from_pdg_id(pdg_id, long_name=False)) == pdg_id
        long_name = val[-1]
        pdg_id = pdg.get_pdg_id_from_name(long_name)
        assert pdg.get_name_from_pdg_id(pdg_id, long_name=True) == long_name
        assert pdg.get_pdg_id_from_name(
            pdg.get_name_from_pdg_id(pdg_id, long_name=True)) == pdg_id
    # Test is_proton, is_lepton, is_ion
    assert pdg.is_proton(2212)
    assert not any([pdg.is_proton(pdg_id) for pdg_id in range(-10000, 10000) if pdg_id != 2212])
    assert not any([pdg.is_proton(10*pdg_id) for pdg_id in range(100000000, 100150000)])
    assert all([pdg.is_lepton(pdg_id) and pdg.is_lepton(-pdg_id) for pdg_id in range(11,17)])
    assert not any([pdg.is_lepton(pdg_id) for pdg_id in range(-10000, 10000) if abs(pdg_id) not in range(11,17)])
    assert not any([pdg.is_lepton(10*pdg_id) for pdg_id in range(100000000, 100150000)])
    def _A_geq_Z(pdg_id):
        tmpid = pdg_id - 1000000000
        L = int(tmpid/1e7)
        tmpid -= L*1e7
        Z = int(tmpid /1e4)
        tmpid -= Z*1e4
        A = int(tmpid /10)
        return A > Z > 0
    assert all([pdg.is_ion(10*pdg_id) if _A_geq_Z(10*pdg_id) else not pdg.is_ion(10*pdg_id)
                for pdg_id in range(100000000, 100150000)])
    assert not any([pdg.is_ion(pdg_id) for pdg_id in range(-10000, 10000)])
    # Test table names
    _names = [
        [['e⁻', 'e', 'electron'], 11],
        [['e⁺', 'positron'], -11],
        [['νₑ', 'electron neutrino'], 12],
        [['μ⁻', 'μ', 'muon-', 'muon'], 13],
        [['μ⁺', 'muon+', 'anti-muon'], -13],
        [['νμ', 'muon neutrino'], 14],
        [['τ⁻', 'τ', 'tau-', 'tau'], 15],
        [['τ⁺', 'tau+', 'anti-tau'], -15],
        [['ντ', 'tau neutrino'], 16],
        [['γ⁰', 'γ', 'photon'], 22],
        [['π⁰', 'π', 'pion', 'pion0', 'pi0'], 111],
        [['π⁺', 'pion+', 'pi+'], 211],
        [['π⁻', 'pion-', 'pi-'], -211],
        [['K⁰', 'kaon', 'kaon0'], 311],
        [['K⁺', 'kaon+'], 321],
        [['K⁻', 'kaon-'], -321],
        [['KL', 'long kaon'], 130],
        [['Kₛ', 'short kaon'], 310],
        [['D⁰', 'D'], 421],
        [['D⁺'], 411],
        [['D⁻'], -411],
        [['Dₛ⁺'], 431],
        [['Dₛ⁻'], -431],
        [['p⁺', 'p', 'proton'], 2212],
        [[ 'p⁻', 'anti-proton'], -2212],
        [['n⁰', 'n', 'neutron'], 2112],
        [['Δ⁺⁺', 'delta++'], 2224],
        [['Δ⁺', 'delta+'], 2214],
        [['Δ⁰', 'delta0'], 2114],
        [['Δ⁻', 'delta-'], 1114],
        [['Λ⁰', 'Λ', 'lambda'], 3122],
        [['Λc⁺', 'lambdac+'], 4122],
        [['Σ⁺', 'sigma+'], 3222],
        [['Σ⁰', 'Σ', 'sigma', 'sigma0'], 3212],
        [['Σ⁻', 'sigma-'], 3112],
        [['Ξ⁰', 'Ξ', 'xi', 'xi0'], 3322],
        [['Ξ⁻', 'xi-'], 3312],
        [['Ξc⁰', 'Ξc', 'xic', 'xic0'], 4132],
        [['Ξc⁺', 'xic+'], 4232],
        [["Ξ'c⁰", "Ξ'c", "xiprimec", "xiprimec0"], 4312],
        [["Ξ'c⁺", "xiprimec+"], 4322],
        [['Ω⁻', 'omega-'], 3334],
        [['Ωc⁰', 'Ωc', 'omegac', 'omegac0'], 4332],
        [['²H', 'H2', 'hydrogen-2', 'deuteron'], 1000010020],
        [['³H', 'H3', 'hydrogen-3', 'triton'], 1000010030]
    ]
    for names, pdg_id in _names:
        for name in names:
            name_variants = [[nn, nn.lower(), nn.upper(), nn.capitalize()]
                             for nn in [name, pdg._to_normal_script(name),
                                        pdg._digits_to_superscript(name)]]
            name_variants = [nnn for nn in name_variants for nnn in nn]
            pdg_ids = np.unique(pdg.get_pdg_id_from_name(name_variants))
            assert len(pdg_ids) == 1
            assert pdg_ids[0] == pdg_id
            antiname_variants = [[f'anti{pdg._flip_end_sign(nn)}',
                                  f'anti-{pdg._flip_end_sign(nn)}',
                                  f'anti {pdg._flip_end_sign(nn)}']
                                 for nn in name_variants]
            antiname_variants = [nnn for nn in antiname_variants for nnn in nn]
            pdg_ids = np.unique(pdg.get_pdg_id_from_name(antiname_variants))
            assert len(pdg_ids) == 1
            assert pdg_ids[0] == -pdg_id
    # Test ion names
    for Z, name in pdg.elements_long.items():
        assert name == pdg.get_element_full_name_from_Z(Z)
        for A in range(Z+1, 4*Z):
            if Z == 1 and A < 4:
                long_name = 'deuteron' if A == 2 else 'triton'
            else:
                long_name = f'{name}-{A}'
            pdg_id = pdg.get_pdg_id_ion(A, Z)
            assert pdg.get_pdg_id_from_name(f'{name}{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name} {A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}-{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}_{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}.{A}') == pdg_id
            assert pdg.get_name_from_pdg_id(pdg_id, long_name=True) == long_name
            assert pdg.get_name_from_pdg_id(
                        pdg.get_pdg_id_from_name(long_name),
                        long_name=True) == long_name
            assert pdg.get_pdg_id_from_name(
                        pdg.get_name_from_pdg_id(pdg_id,
                        long_name=True)) == pdg_id
    for Z, name in pdg.elements.items():
        assert name == pdg.get_element_name_from_Z(Z)
        for A in range(Z+1, 4*Z):
            short_name = f'{pdg._digits_to_superscript(A)}{name}'
            pdg_id = pdg.get_pdg_id_ion(A, Z)
            assert pdg.get_pdg_id_from_name(f'{A}{name}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name} {A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}-{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}_{A}') == pdg_id
            assert pdg.get_pdg_id_from_name(f'{name}.{A}') == pdg_id
            assert pdg.get_name_from_pdg_id(pdg_id, long_name=False) == short_name
            assert pdg.get_name_from_pdg_id(
                        pdg.get_pdg_id_from_name(short_name),
                        long_name=False) == short_name
            assert pdg.get_pdg_id_from_name(
                        pdg.get_name_from_pdg_id(pdg_id,
                        long_name=False)) == pdg_id


def test_charge_and_properties():
    for pdg_id, vals in pdg.pdg_table.items():
        q, _, _, _ = pdg.get_properties_from_pdg_id(pdg_id)
        assert q == vals[0]
        anti_q, _, _, _ = pdg.get_properties_from_pdg_id(-pdg_id)
        if vals[1].endswith('⁺') or vals[1].endswith('+'):
            assert np.sign(q) == 1
            assert np.sign(anti_q) == -1
        elif vals[1].endswith('⁻') or vals[1].endswith('-'):
            assert np.sign(q) == -1
            assert np.sign(anti_q) == 1
        elif vals[1].endswith('⁰') or vals[1].endswith('0'):
            assert np.sign(q) == 0
            assert np.sign(anti_q) == 0
        else:
            # Particle names without a charge in the first column of the PDG table
            # should be neutral (except for Hydrogen isotopes)
            if 0 < abs(pdg_id) < 1000000000:
                assert np.sign(q) == 0
                assert np.sign(anti_q) == 0
    for pdg_id in range(100000000, 100119000):
        pdg_id *= 10
        if pdg.is_ion(pdg_id):
            q, A, Z, name = pdg.get_properties_from_pdg_id(pdg_id, long_name=False, subscripts=False)
            assert A > Z > 0
            assert q > 0
            assert q == Z
            name = name.replace(f"{A}", '')
            assert pdg.get_Z_from_element_name(name) == Z
            assert name == pdg.get_element_name_from_Z(Z)
            q, A, Z, name = pdg.get_properties_from_pdg_id(pdg_id, long_name=True)
            if name == 'deuteron':
                assert A == 2
                assert Z == 1
                assert q == 1
            elif name == 'triton':
                assert A == 3
                assert Z == 1
                assert q == 1
            else:
                assert A > Z > 0
                assert q > 0
                assert q == Z
                name_parts = name.split('-')
                assert pdg.get_Z_from_element_name(name_parts[0]) == Z
                assert name_parts[1] == f"{A}"
                assert name_parts[0] == pdg.get_element_full_name_from_Z(Z)


def test_masses():
    assert pdg.get_pdg_id_from_mass_charge(0, 0) == 22
    assert pdg.get_pdg_id_from_mass_charge(511e3, -1) == 11
    assert pdg.get_pdg_id_from_mass_charge(511e3, 1) == -11
    assert pdg.get_pdg_id_from_mass_charge(105.7e6, -1) == 13
    assert pdg.get_pdg_id_from_mass_charge(105.7e6, 1) == -13
    assert pdg.get_pdg_id_from_mass_charge(135.0e6, 0) == 111
    assert pdg.get_pdg_id_from_mass_charge(139.6e6, 1) == 211
    assert pdg.get_pdg_id_from_mass_charge(139.6e6, -1) == -211
    assert pdg.get_pdg_id_from_mass_charge(497.6e6, 0) == 311
    assert pdg.get_pdg_id_from_mass_charge(493.7e6, 1) == 321
    assert pdg.get_pdg_id_from_mass_charge(493.7e6, -1) == -321
    assert pdg.get_pdg_id_from_mass_charge(938.3e6, 1) == 2212
    assert pdg.get_pdg_id_from_mass_charge(938.3e6, -1) == -2212
    assert pdg.get_pdg_id_from_mass_charge(939.6e6, 0) == 2112
    assert pdg.get_pdg_id_from_mass_charge(1875.6e6, 1) == 1000010020
    assert pdg.get_pdg_id_from_mass_charge(2808.9e6, 1) == 1000010030
    assert np.isclose(pdg.get_mass_from_pdg_id(22), 0, rtol=1e-4)
    assert np.isclose(pdg.get_mass_from_pdg_id(11), 511e3, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(-11), 511e3, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(13), 105.7e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(-13), 105.7e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(111), 135.0e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(211), 139.6e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(-211), 139.6e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(311), 497.6e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(321), 493.7e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(-321), 493.7e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(2212), 938.3e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(-2212), 938.3e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(2112), 939.6e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(1000010020), 1875.6e6, rtol=1e-3)
    assert np.isclose(pdg.get_mass_from_pdg_id(1000010030), 2808.9e6, rtol=1e-3)
    # Test the ion masses defined in the mass table
    for massdef, mass in masses.__dict__.items():
        if massdef.endswith('_MASS_EV'):
            massdef = massdef[:-8]
            if any([i.isdigit() for i in massdef]):
                A = int(''.join([i for i in massdef if i.isdigit()]))
                el = massdef.replace(f'{A}', '')
                if el in pdg.elements:
                    Z = pdg.get_Z_from_element_name(el)
                    pdg_id = pdg.get_pdg_id_ion(A, Z)
                    assert pdg.get_pdg_id_from_mass_charge(mass, Z) == pdg_id
                    assert assertnp.isclose(pdg.get_mass_from_pdg_id(pdg_id), mass, rtol=1e-7)
    for Z in pdg.elements.keys():
        if Z < 3 or Z == 6 or Z==26:
            rtol = 1e-2
        else:
            rtol = 1e-3
        for A in range(Z+1, 4*Z):
            pdg_id = pdg.get_pdg_id_ion(A, Z)
            assert pdg.get_pdg_id_from_mass_charge(A*masses.U_MASS_EV, Z) == pdg_id
            assert np.isclose(pdg.get_mass_from_pdg_id(pdg_id), A*masses.U_MASS_EV, rtol=rtol)


def test_lead_208():
    pdg_id = 1000822080
    assert pdg.get_pdg_id_ion(208, 82) == pdg_id
    assert pdg.get_pdg_id_from_mass_charge(masses.Pb208_MASS_EV, 82) == pdg_id
    assert pdg.get_name_from_pdg_id(pdg_id, long_name=False) == '²⁰⁸Pb'
    assert pdg.get_name_from_pdg_id(pdg_id, long_name=True) == 'Lead-208'
    assert pdg.get_pdg_id_from_name('Pb208')    == pdg_id
    assert pdg.get_pdg_id_from_name('Pb 208')   == pdg_id
    assert pdg.get_pdg_id_from_name('Pb-208')   == pdg_id
    assert pdg.get_pdg_id_from_name('Pb_208')   == pdg_id
    assert pdg.get_pdg_id_from_name('Pb.208')   == pdg_id
    assert pdg.get_pdg_id_from_name('pb208')    == pdg_id
    assert pdg.get_pdg_id_from_name('208pb')    == pdg_id
    assert pdg.get_pdg_id_from_name('208Pb')    == pdg_id
    assert pdg.get_pdg_id_from_name('lead208')  == pdg_id
    assert pdg.get_pdg_id_from_name('Lead 208') == pdg_id
    assert pdg.get_pdg_id_from_name('Lead_208') == pdg_id
    assert pdg._mass_consistent(pdg_id, masses.Pb208_MASS_EV)
    assert pdg.get_element_name_from_Z(82) == 'Pb'
    assert pdg.get_element_full_name_from_Z(82) == 'Lead'
    xo.assert_allclose(pdg.get_mass_from_pdg_id(pdg_id), masses.Pb208_MASS_EV,
                       rtol=1e-10, atol=0)
    assert pdg.get_properties_from_pdg_id(pdg_id) == (82., 208, 82, '²⁰⁸Pb')


@for_all_test_contexts
def test_build_reference_from_pdg_id(test_context):
    particle_ref_proton  = xt.particles.reference_from_pdg_id(pdg_id='proton',
                                                    _context=test_context)
    particle_ref_proton.move(_context=xo.context_default)
    assert particle_ref_proton.pdg_id == 2212
    particle_ref_lead = xt.particles.reference_from_pdg_id(pdg_id='Pb208',
                                                 _context=test_context)
    particle_ref_lead.move(_context=xo.context_default)
    xo.assert_allclose(particle_ref_lead.q0, 82.)
    xo.assert_allclose(particle_ref_lead.mass0, masses.Pb208_MASS_EV)



import numpy as np

from cpymad.madx import Madx
import xtrack as xt
import xdeps as xd

mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use(sequence="lhcb1")

line = xt.Line.from_madx_sequence(mad.sequence['lhcb1'], deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1,
                                 gamma0=mad.sequence.lhcb1.beam.gamma)
line.build_tracker()

line.functions['ramp_on_x1'] = xd.FunctionPieceWiseLinear(x=[0.5, 1, 3, 5],
                                                          y=[0, 100, 100, 0])
line.vars['t_turn_s'] = 0
line.vars['on_x1'] = line.functions.ramp_on_x1(line.vars['t_turn_s'])

assert np.isclose(line.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 1
assert np.isclose(line.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 2
assert np.isclose(line.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 3
assert np.isclose(line.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 4
assert np.isclose(line.twiss()['px', 'ip1'], 50e-6, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 5
assert np.isclose(line.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)
line.vars['t_turn_s'] = 6
assert np.isclose(line.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)

# Test a knob involving functions
line.vars['on_x8'] = 100
line.vars['phi_ir8'] = 0
assert np.isclose(line.twiss()['px', 'ip8'], 100e-6, atol=1e-7, rtol = 0)
assert np.isclose(line.twiss()['py', 'ip8'], 0, atol=1e-7, rtol = 0)

line.vars['phi_ir8'] = 90
assert np.isclose(line.twiss()['px', 'ip8'], 0, atol=1e-7, rtol = 0)
assert np.isclose(line.twiss()['py', 'ip8'], 100e-6, atol=1e-7, rtol = 0)

line.vars['phi_ir8'] = 0

# Test json roundtrip
line.to_json('test.json')
line2 = xt.Line.from_json('test.json')
line2.build_tracker()

assert np.isclose(line2.twiss()['px', 'ip8'], 100e-6, atol=1e-7, rtol = 0)
assert np.isclose(line2.twiss()['py', 'ip8'], 0, atol=1e-7, rtol = 0)
line2.vars['phi_ir8'] = 90
assert np.isclose(line2.twiss()['px', 'ip8'], 0, atol=1e-7, rtol = 0)
assert np.isclose(line2.twiss()['py', 'ip8'], 100e-6, atol=1e-7, rtol = 0)

assert np.isclose(line2.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 1
assert np.isclose(line2.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 2
assert np.isclose(line2.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 3
assert np.isclose(line2.twiss()['px', 'ip1'], 100e-6, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 4
assert np.isclose(line2.twiss()['px', 'ip1'], 50e-6, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 5
assert np.isclose(line2.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)
line2.vars['t_turn_s'] = 6
assert np.isclose(line2.twiss()['px', 'ip1'], 0, atol=1e-7, rtol = 0)


# Remember to test multiline2




import json
import numpy as np

import xtrack as xt
import xpart as xp
import xobjects as xo
from cpymad.madx import Madx

mad = Madx()
mad.call('SPS_2021_Pb_ions_thin_test.seq')
mad.use('sps')
mad.twiss()

twmad_4d = mad.table.twiss.dframe()
summad_4d = mad.table.summ.dframe()

V_RF = 1.7e6 # V (control room definition, energy gain per charge)

# I switch on one cavity
charge = mad.sequence.sps.beam.charge
mad.sequence.sps.elements['actcse.31632'].volt = V_RF/1e6 * charge
mad.sequence.sps.elements['actcse.31632'].lag = 0
mad.sequence.sps.elements['actcse.31632'].freq = 200.

twmad_6d = mad.table.twiss.dframe()
summad_6d = mad.table.summ.dframe()

mad.emit()
qs_mad = mad.table.emitsumm.qs[0]

# Make xsuite line and tracker
line = xt.Line.from_madx_sequence(mad.sequence.sps, deferred_expressions=True)
line.particle_ref = xp.Particles(mass0=mad.sequence.sps.beam.mass*1e9,
                                 q0=mad.sequence.sps.beam.charge,
                                 gamma0=mad.sequence.sps.beam.gamma)
tracker = line.build_tracker()

tw = line.twiss()

assert np.isclose(tw.qs, qs_mad, atol=1e-6)
assert np.isclose(tw.qx, summad_4d.q1, atol=1e-5)
assert np.isclose(tw.qy, summad_4d.q2, atol=1e-5)
assert np.isclose(tw.dqx, summad_6d.dq1, atol=0.5)
assert np.isclose(tw.dqy, summad_6d.dq2, atol=0.5)

with open('line_and_particle.json', 'w') as f:
    json.dump(line.to_dict(), f, cls=xo.JEncoder, indent=4)




import xtrack as xt
import xobjects as xo
import numpy as np
from xobjects.test_helpers import for_all_test_contexts

assert_allclose = xo.assert_allclose

@for_all_test_contexts
def test_slice_thin_and_insert_with_replicas(test_context):

    elements = {
        'e0': xt.Bend(k0=0.3, h=0.31, length=1),
        'e1': xt.Replica(parent_name='e0'),
        'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e3': xt.Replica(parent_name='e2'),
        'e4': xt.Replica(parent_name='e3'), # Replica of replica
    }
    line = xt.Line(elements=elements,
                element_names=list(elements.keys()))
    line.build_tracker(_context=test_context)

    element_no_repl={
        'e0': xt.Bend(k0=0.3, h=0.31, length=1),
        'e1': xt.Bend(k0=0.3, h=0.31, length=1),
        'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e3': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e4': xt.Bend(k0=-0.4, h=-0.41, length=1),
    }

    line_no_repl = xt.Line(elements=element_no_repl,
                        element_names=list(element_no_repl.keys()))
    line_no_repl.build_tracker(_context=test_context)

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()

    assert np.all(tt.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])
    assert np.all(tt_no_repl.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])

    assert np.all(tt.parent_name == [None, 'e0', None, 'e2', 'e3', None])
    assert np.all(tt_no_repl.parent_name == [None, None, None, None, None, None])

    assert_allclose(tt.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)
    assert_allclose(tt_no_repl.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)

    p0 = xt.Particles(p0c=1e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                      _context=test_context)
    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(xt.Teapot(2, mode='thin'), name='e2|e3|e4')])
    line.build_tracker(_context=test_context)

    line_no_repl.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(xt.Teapot(2, mode='thin'), name='e2|e3|e4')])
    line_no_repl.build_tracker(_context=test_context)

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()

    assert np.all(tt.name == np.array(['e0', 'e1', 'e2_entry', 'e2..entry_map', 'drift_e2..0', 'e2..0',
        'drift_e2..1', 'e2..1', 'drift_e2..2', 'e2..exit_map', 'e2_exit',
        'e3_entry', 'e3..entry_map', 'drift_e3..0', 'e3..0', 'drift_e3..1',
        'e3..1', 'drift_e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'drift_e4..0', 'e4..0', 'drift_e4..1', 'e4..1',
        'drift_e4..2', 'e4..exit_map', 'e4_exit', '_end_point']))

    assert np.all(tt_no_repl.name == np.array([
        'e0', 'e1', 'e2_entry', 'e2..entry_map', 'drift_e2..0', 'e2..0',
        'drift_e2..1', 'e2..1', 'drift_e2..2', 'e2..exit_map', 'e2_exit',
        'e3_entry', 'e3..entry_map', 'drift_e3..0', 'e3..0', 'drift_e3..1',
        'e3..1', 'drift_e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'drift_e4..0', 'e4..0', 'drift_e4..1', 'e4..1',
        'drift_e4..2', 'e4..exit_map', 'e4_exit', '_end_point']))

    assert np.all(tt.parent_name == np.array([
        None, 'e0', None, 'e2', 'e2', 'e2', 'e2', 'e2', 'e2', 'e2', None,
        None, 'e2', 'e2', 'e2', 'e2', 'e2', 'e2', 'e2', None, None, 'e2',
        'e2', 'e2', 'e2', 'e2', 'e2', 'e2', None, None]))
    assert np.all(tt_no_repl.parent_name == np.array([None, None, None,
        'e2', 'e2', 'e2', 'e2', 'e2', 'e2', 'e2', None,
        None, 'e3', 'e3', 'e3', 'e3', 'e3', 'e3', 'e3', None, None, 'e4',
        'e4', 'e4', 'e4', 'e4', 'e4', 'e4', None, None]))

    assert_allclose(tt.s, np.array([
        0.        , 1.        , 2.        , 2.        , 2.        ,
        2.16666667, 2.16666667, 2.83333333, 2.83333333, 3.        ,
        3.        , 3.        , 3.        , 3.        , 3.16666667,
        3.16666667, 3.83333333, 3.83333333, 4.        , 4.        ,
        4.        , 4.        , 4.        , 4.16666667, 4.16666667,
        4.83333333, 4.83333333, 5.        , 5.        , 5.        ]))

    assert_allclose(tt_no_repl.s, np.array([
        0.        , 1.        , 2.        , 2.        , 2.        ,
        2.16666667, 2.16666667, 2.83333333, 2.83333333, 3.        ,
        3.        , 3.        , 3.        , 3.        , 3.16666667,
        3.16666667, 3.83333333, 3.83333333, 4.        , 4.        ,
        4.        , 4.        , 4.        , 4.16666667, 4.16666667,
        4.83333333, 4.83333333, 5.        , 5.        , 5.        ]))


    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

    line.discard_tracker()
    line.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
    line.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
    line.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
    line.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
    line.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
    line.build_tracker(_context=test_context)

    line_no_repl.discard_tracker()
    line_no_repl.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
    line_no_repl.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
    line_no_repl.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
    line_no_repl.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
    line_no_repl.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
    line_no_repl.build_tracker(_context=test_context)

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()


    assert np.all(tt.name == np.array(['e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
        'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
        'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
        'e2..entry_map', 'drift_e2..0', 'e2..0', 'drift_e2..1..0',
        'mkins3', 'drift_e2..1..1', 'e2..1', 'drift_e2..2', 'e2..exit_map',
        'e2_exit', 'e3_entry', 'e3..entry_map', 'drift_e3..0', 'e3..0',
        'drift_e3..1..0', 'mkins4', 'drift_e3..1..1', 'e3..1',
        'drift_e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'drift_e4..0', 'e4..0', 'drift_e4..1..0',
        'mkins5', 'drift_e4..1..1', 'e4..1', 'drift_e4..2', 'e4..exit_map',
        'e4_exit', '_end_point']))

    assert np.all(tt_no_repl.name == np.array(['e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
        'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
        'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
        'e2..entry_map', 'drift_e2..0', 'e2..0', 'drift_e2..1..0',
        'mkins3', 'drift_e2..1..1', 'e2..1', 'drift_e2..2', 'e2..exit_map',
        'e2_exit', 'e3_entry', 'e3..entry_map', 'drift_e3..0', 'e3..0',
        'drift_e3..1..0', 'mkins4', 'drift_e3..1..1', 'e3..1',
        'drift_e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'drift_e4..0', 'e4..0', 'drift_e4..1..0',
        'mkins5', 'drift_e4..1..1', 'e4..1', 'drift_e4..2', 'e4..exit_map',
        'e4_exit', '_end_point']))

    assert np.all(tt.parent_name == np.array([
        None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e0', 'e0', None,
        'e0', 'e0', None, None, 'e2', 'e2', 'e2', 'e2', None, 'e2', 'e2',
        'e2', 'e2', None, None, 'e2', 'e2', 'e2', 'e2', None, 'e2', 'e2',
        'e2', 'e2', None, None, 'e2', 'e2', 'e2', 'e2', None, 'e2', 'e2',
        'e2', 'e2', None, None]))

    assert np.all(tt_no_repl.parent_name == np.array([
        None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e1', 'e1', None,
        'e1', 'e1', None, None, 'e2', 'e2', 'e2', 'e2', None, 'e2', 'e2',
        'e2', 'e2', None, None, 'e3', 'e3', 'e3', 'e3', None, 'e3', 'e3',
        'e3', 'e3', None, None, 'e4', 'e4', 'e4', 'e4', None, 'e4', 'e4',
        'e4', 'e4', None, None]))

    assert_allclose(tt.s, [
        0.        , 0.        , 0.        , 0.5       , 0.5       ,
        1.        , 1.        , 1.        , 1.        , 1.        ,
        1.5       , 1.5       , 2.        , 2.        , 2.        ,
        2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
        2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
        3.        , 3.        , 3.        , 3.16666667, 3.16666667,
        3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
        5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

    assert_allclose(tt_no_repl.s, [
        0.        , 0.        , 0.        , 0.5       , 0.5       ,
        1.        , 1.        , 1.        , 1.        , 1.        ,
        1.5       , 1.5       , 2.        , 2.        , 2.        ,
        2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
        2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
        3.        , 3.        , 3.        , 3.16666667, 3.16666667,
        3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
        5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

@for_all_test_contexts
def test_slice_thick_and_insert_with_replicas(test_context):

    elements = {
        'e0': xt.Bend(k0=0.3, h=0.31, length=1),
        'e1': xt.Replica(parent_name='e0'),
        'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e3': xt.Replica(parent_name='e2'),
        'e4': xt.Replica(parent_name='e3'), # Replica of replica
    }
    line = xt.Line(elements=elements,
                element_names=list(elements.keys()))
    line.build_tracker(_context=test_context)

    assert line['e2']._movable

    element_no_repl={
        'e0': xt.Bend(k0=0.3, h=0.31, length=1),
        'e1': xt.Bend(k0=0.3, h=0.31, length=1),
        'e2': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e3': xt.Bend(k0=-0.4, h=-0.41, length=1),
        'e4': xt.Bend(k0=-0.4, h=-0.41, length=1),
    }

    line_no_repl = xt.Line(elements=element_no_repl,
                        element_names=list(element_no_repl.keys()))
    line_no_repl.build_tracker(_context=test_context)

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()

    assert np.all(tt.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])
    assert np.all(tt_no_repl.name == ['e0', 'e1', 'e2', 'e3', 'e4', '_end_point'])

    assert np.all(tt.parent_name == [None, 'e0', None, 'e2', 'e3', None])
    assert np.all(tt_no_repl.parent_name == [None, None, None, None, None, None])

    assert_allclose(tt.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)
    assert_allclose(tt_no_repl.s, [0., 1., 2., 3., 4., 5.], rtol=0, atol=1e-14)

    p0 = xt.Particles(p0c=1e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                        _context=test_context)
    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

    assert line['e2']._movable
    # line['e2']._mark = True

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(xt.Teapot(3, mode='thick'), name='e2|e3|e4')])
    assert line['e2']._movable
    line.build_tracker(_context=test_context)

    assert line['e2']._movable

    line_no_repl.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(xt.Teapot(3, mode='thick'), name='e2|e3|e4')])
    line_no_repl.build_tracker(_context=test_context)

    assert line['e2']._movable

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()

    assert np.all(tt.name == np.array([
        'e0', 'e1', 'e2_entry', 'e2..entry_map', 'e2..0', 'e2..1', 'e2..2',
        'e2..exit_map', 'e2_exit', 'e3_entry', 'e3..entry_map', 'e3..0',
        'e3..1', 'e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'e4..0', 'e4..1', 'e4..2', 'e4..exit_map',
        'e4_exit', '_end_point']))

    assert np.all(tt_no_repl.name == np.array([
        'e0', 'e1', 'e2_entry', 'e2..entry_map', 'e2..0', 'e2..1', 'e2..2',
        'e2..exit_map', 'e2_exit', 'e3_entry', 'e3..entry_map', 'e3..0',
        'e3..1', 'e3..2', 'e3..exit_map', 'e3_exit', 'e4_entry',
        'e4..entry_map', 'e4..0', 'e4..1', 'e4..2', 'e4..exit_map',
        'e4_exit', '_end_point']))

    assert np.all(tt.parent_name == np.array([
        None, 'e0', None, 'e2', 'e2', 'e2', 'e2', 'e2', None, None, 'e2',
        'e2', 'e2', 'e2', 'e2', None, None, 'e2', 'e2', 'e2', 'e2', 'e2',
        None, None]))
    assert np.all(tt_no_repl.parent_name == np.array([
        None, None, None, 'e2', 'e2', 'e2', 'e2', 'e2', None, None, 'e3',
        'e3', 'e3', 'e3', 'e3', None, None, 'e4', 'e4', 'e4', 'e4', 'e4',
        None, None]))

    assert_allclose(tt.s, np.array([
        0.        , 1.        , 2.        , 2.        , 2.        ,
        2.16666667, 2.83333333, 3.        , 3.        , 3.        ,
        3.        , 3.        , 3.16666667, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.83333333, 5.        , 5.        , 5.        ]))

    assert_allclose(tt_no_repl.s, np.array([
        0.        , 1.        , 2.        , 2.        , 2.        ,
        2.16666667, 2.83333333, 3.        , 3.        , 3.        ,
        3.        , 3.        , 3.16666667, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.83333333, 5.        , 5.        , 5.        ]))


    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)

    assert line['e2']._movable

    line.discard_tracker()
    assert line['e2']._movable
    line.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
    line.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
    line.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
    line.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
    line.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
    line.build_tracker(_context=test_context)
    assert line['e2']._movable

    line_no_repl.discard_tracker()
    line_no_repl.insert_element(name='mkins1', element=xt.Marker(), at_s=0.5)
    line_no_repl.insert_element(name='mkins2', element=xt.Marker(), at_s=1.5)
    line_no_repl.insert_element(name='mkins3', element=xt.Marker(), at_s=2.5)
    line_no_repl.insert_element(name='mkins4', element=xt.Marker(), at_s=3.5)
    line_no_repl.insert_element(name='mkins5', element=xt.Marker(), at_s=4.5)
    line_no_repl.build_tracker(_context=test_context)
    assert line['e2']._movable

    tt = line.get_table()
    tt_no_repl = line_no_repl.get_table()


    assert np.all(tt.name == np.array([
        'e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
        'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
        'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
        'e2..entry_map', 'e2..0', 'e2..1_entry', 'e2..1..0', 'mkins3',
        'e2..1..1', 'e2..1_exit', 'e2..2', 'e2..exit_map', 'e2_exit',
        'e3_entry', 'e3..entry_map', 'e3..0', 'e3..1_entry', 'e3..1..0',
        'mkins4', 'e3..1..1', 'e3..1_exit', 'e3..2', 'e3..exit_map',
        'e3_exit', 'e4_entry', 'e4..entry_map', 'e4..0', 'e4..1_entry',
        'e4..1..0', 'mkins5', 'e4..1..1', 'e4..1_exit', 'e4..2',
        'e4..exit_map', 'e4_exit', '_end_point']))

    assert np.all(tt_no_repl.name == np.array([
        'e0_entry', 'e0..entry_map', 'e0..0', 'mkins1', 'e0..1',
        'e0..exit_map', 'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0',
        'mkins2', 'e1..1', 'e1..exit_map', 'e1_exit', 'e2_entry',
        'e2..entry_map', 'e2..0', 'e2..1_entry', 'e2..1..0', 'mkins3',
        'e2..1..1', 'e2..1_exit', 'e2..2', 'e2..exit_map', 'e2_exit',
        'e3_entry', 'e3..entry_map', 'e3..0', 'e3..1_entry', 'e3..1..0',
        'mkins4', 'e3..1..1', 'e3..1_exit', 'e3..2', 'e3..exit_map',
        'e3_exit', 'e4_entry', 'e4..entry_map', 'e4..0', 'e4..1_entry',
        'e4..1..0', 'mkins5', 'e4..1..1', 'e4..1_exit', 'e4..2',
        'e4..exit_map', 'e4_exit', '_end_point']))

    assert np.all(tt.parent_name == np.array([
        None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e0', 'e0', None,
        'e0', 'e0', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
        'e2', 'e2', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
        'e2', 'e2', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
        'e2', 'e2', None, None]))

    assert np.all(tt_no_repl.parent_name == np.array([
        None, 'e0', 'e0', None, 'e0', 'e0', None, None, 'e1', 'e1', None,
        'e1', 'e1', None, None, 'e2', 'e2', None, 'e2', None, 'e2', None,
        'e2', 'e2', None, None, 'e3', 'e3', None, 'e3', None, 'e3', None,
        'e3', 'e3', None, None, 'e4', 'e4', None, 'e4', None, 'e4', None,
        'e4', 'e4', None, None]))

    assert_allclose(tt.s, [
        0.        , 0.        , 0.        , 0.5       , 0.5       ,
        1.        , 1.        , 1.        , 1.        , 1.        ,
        1.5       , 1.5       , 2.        , 2.        , 2.        ,
        2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
        2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
        3.        , 3.        , 3.        , 3.16666667, 3.16666667,
        3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
        5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

    assert_allclose(tt_no_repl.s, [
        0.        , 0.        , 0.        , 0.5       , 0.5       ,
        1.        , 1.        , 1.        , 1.        , 1.        ,
        1.5       , 1.5       , 2.        , 2.        , 2.        ,
        2.        , 2.        , 2.16666667, 2.16666667, 2.5       ,
        2.5       , 2.83333333, 2.83333333, 3.        , 3.        ,
        3.        , 3.        , 3.        , 3.16666667, 3.16666667,
        3.5       , 3.5       , 3.83333333, 3.83333333, 4.        ,
        4.        , 4.        , 4.        , 4.        , 4.16666667,
        4.16666667, 4.5       , 4.5       , 4.83333333, 4.83333333,
        5.        , 5.        , 5.        ], rtol=0, atol=5e-9)

    p1 = p0.copy()
    p2 = p0.copy()

    line.track(p1)
    line_no_repl.track(p2)
    assert line['e2']._movable

    assert_allclose(p2.x, p1.x, rtol=0, atol=1e-14)
    assert_allclose(p2.px, p1.px, rtol=0, atol=1e-14)
    assert_allclose(p2.y, p1.y, rtol=0, atol=1e-14)
    assert_allclose(p2.py, p1.py, rtol=0, atol=1e-14)
    assert_allclose(p2.zeta, p1.zeta, rtol=0, atol=1e-14)
    assert_allclose(p2.delta, p1.delta, rtol=0, atol=1e-14)



import numpy as np
import xtrack as xt
import xobjects as xo

env = xt.Environment()

components = [
    env.new('q1', 'Quadrupole', length=2.0, anchor='start', at=1.),
    env.new('q2', 'q1', anchor='start', at=10., from_='q1', from_anchor='end'),
    env.new('s2', 'Sextupole', length=0.1, anchor='end', at=-1., from_='q2', from_anchor='start'),

    env.new('q3', 'Quadrupole', length=2.0, at=20.),
    env.new('q4', 'q3', anchor='start', at=0., from_='q3', from_anchor='end'),
    env.new('q5', 'q3'),

    # Sandwich of markers expected [m2.0, m2, m2.1.0, m2.1.1. m2.1]
    env.new('m2_0', 'Marker', at=0., from_='m2', from_anchor='start'),
    env.new('m2', 'Marker', at=0., from_='q2', from_anchor='start'),
    env.new('m2_1', 'Marker', at=0., from_='m2', from_anchor='end'),
    env.new('m2_1_0', 'Marker', at=0., from_='m2_1', from_anchor='start'),
    env.new('m2_1_1', 'Marker'),

    env.new('m1', 'Marker', at=0., from_='q1', from_anchor='start'),

    env.new('m4', 'Marker', at=0., from_='q4', from_anchor='start'),
    env.new('m3', 'Marker', at=0., from_='q3', from_anchor='end'),
]

line = env.new_line(components=components)

tt = line.get_table()

tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'm1', 'q1', 'drift_2', 's2', 'drift_3', 'm2_0', 'm2',
       'm2_1_0', 'm2_1_1', 'm2_1', 'q2', 'drift_4', 'q3', 'm3', 'm4',
       'q4', 'q5', '_end_point']))
xo.assert_allclose(tt.s, np.array(
    [ 0. ,  1. ,  1. ,  3. , 11.9, 12. , 13. , 13. , 13. , 13. , 13. ,
       13. , 15. , 19. , 21. , 21. , 21. , 23. , 25. ]),
    rtol=0., atol=1e-14)
xo.assert_allclose(tt.s_start, np.array(
    [ 0. ,  1. ,  1. ,  3. , 11.9, 12. , 13. , 13. , 13. , 13. , 13. ,
       13. , 15. , 19. , 21. , 21. , 21. , 23. , 25. ]),
    rtol=0., atol=1e-14)
xo.assert_allclose(tt.s_end, np.array(
    [ 1. ,  1. ,  3. , 11.9, 12. , 13. , 13. , 13. , 13. , 13. , 13. ,
       15. , 19. , 21. , 21. , 21. , 23. , 25. , 25. ]),
    rtol=0., atol=1e-14)
xo.assert_allclose(tt.s_center, 0.5*(tt.s_start + tt.s_end),
    rtol=0., atol=1e-14)



import numpy as np
from cpymad.madx import Madx
import xtrack as xt

mad_b2 = Madx()
mad_b2.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad_b2.use(sequence='lhcb2')
twb2mad = mad_b2.twiss()
summb2mad = mad_b2.table.summ

mad_b4 = Madx()
mad_b4.call('../../test_data/hllhc15_noerrors_nobb/sequence_b4.madx')
mad_b4.use(sequence='lhcb2')
twb4mad = mad_b4.twiss()
summb4mad = mad_b4.table.summ

line_b4 = xt.Line.from_madx_sequence(mad_b4.sequence['lhcb2'],
                                     deferred_expressions=True)
line_b4.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7000e9)

line_b4.build_tracker()
twb4xt = line_b4.twiss()

twb2xt = line_b4.twiss().reverse()

# Compute twiss also from W matrix to check the W matrix
Ws = np.array(twb2xt.W_matrix)
betx = Ws[:, 0, 0]**2 + Ws[:, 0, 1]**2
bety = Ws[:, 2, 2]**2 + Ws[:, 2, 3]**2
gamx = Ws[:, 1, 0]**2 + Ws[:, 1, 1]**2
gamy = Ws[:, 3, 2]**2 + Ws[:, 3, 3]**2
alfx = - Ws[:, 0, 0] * Ws[:, 1, 0] - Ws[:, 0, 1] * Ws[:, 1, 1]
alfy = - Ws[:, 2, 2] * Ws[:, 3, 2] - Ws[:, 2, 3] * Ws[:, 3, 3]

import matplotlib.pyplot as plt

plt.close('all')

fig1 = plt.figure(1)
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(twb2mad['s'], twb2mad['betx'], 'b')
spbet.plot(twb2xt['s'], twb2xt['betx'], '--', color='lightblue')
spbet.plot(twb2mad['s'], twb2mad['bety'], 'r')
spbet.plot(twb2xt['s'], twb2xt['bety'], '--', color='darkred')

spco.plot(twb2mad['s'], twb2mad['x'], 'b')
spco.plot(twb2xt['s'], twb2xt['x'], '--', color='lightblue')
spco.plot(twb2mad['s'], twb2mad['y'], 'r')
spco.plot(twb2xt['s'], twb2xt['y'], '--', color='darkred')

spdisp.plot(twb2mad['s'], twb2mad['dx'], 'b')
spdisp.plot(twb2xt['s'], twb2xt['dx'], '--', color='lightblue')
spdisp.plot(twb2mad['s'], twb2mad['dy'], 'r')
spdisp.plot(twb2xt['s'], twb2xt['dy'], '--', color='darkred')

assert np.isclose(summb2mad.q1[0], twb2xt['qx'], rtol=1e-4)
assert np.isclose(summb2mad.q2[0], twb2xt['qy'], rtol=1e-4)
assert np.isclose(summb2mad.dq1, twb2xt['dqx'], atol=0.1, rtol=0)
assert np.isclose(summb2mad.dq2, twb2xt['dqy'], atol=0.1, rtol=0)
assert np.isclose(summb2mad.alfa[0],
    twb2xt['momentum_compaction_factor'], atol=7e-8, rtol=0)
assert np.isclose(twb2xt['qs'], 0.0021, atol=1e-4, rtol=0)

for name in ['mb.b19r5.b2', 'mb.b19r1.b2', 'ip1', 'ip2', 'ip5', 'ip8',
             'mbxf.4l1', 'mbxf.4l5']:
    imad = list(twb2mad['name']).index(name+':1')
    ixt = list(twb2xt['name']).index(name) + 1 # MAD measures at exit

    for sp in [spbet, spco, spdisp]:
        sp.axvline(x=twb2xt['s'][imad])

    assert np.isclose(twb2xt['betx'][ixt], twb2mad['betx'][imad],
                      atol=0, rtol=3e-4)
    assert np.isclose(betx[ixt], twb2mad['betx'][imad], atol=0, rtol=3e-4)
    assert np.isclose(twb2xt['bety'][ixt], twb2mad['bety'][imad],
                      atol=0, rtol=3e-4)
    assert np.isclose(bety[ixt], twb2mad['bety'][imad], atol=0, rtol=3e-4) 
    assert np.isclose(twb2xt['alfx'][ixt], twb2mad['alfx'][imad],
                      atol=1e-1, rtol=0)
    assert np.isclose(alfx[ixt], twb2mad['alfx'][imad], atol=1e-1, rtol=0)
    assert np.isclose(twb2xt['alfy'][ixt], twb2mad['alfy'][imad],
                      atol=1e-1, rtol=0)
    assert np.isclose(alfy[ixt], twb2mad['alfy'][imad], atol=1e-1, rtol=0)
    assert np.isclose(twb2xt['dx'][ixt], twb2mad['dx'][imad], atol=1e-2)
    assert np.isclose(twb2xt['dy'][ixt], twb2mad['dy'][imad], atol=1e-2)
    assert np.isclose(twb2xt['dpx'][ixt], twb2mad['dpx'][imad], atol=3e-4)
    assert np.isclose(twb2xt['dpy'][ixt], twb2mad['dpy'][imad], atol=3e-4)

    assert np.isclose(twb2xt['s'][ixt], twb2mad['s'][imad], atol=5e-6)
    assert np.isclose(twb2xt['x'][ixt], twb2mad['x'][imad], atol=5e-6)
    assert np.isclose(twb2xt['y'][ixt], twb2mad['y'][imad], atol=5e-6)
    assert np.isclose(twb2xt['px'][ixt], twb2mad['px'][imad], atol=1e-7)
    assert np.isclose(twb2xt['py'][ixt], twb2mad['py'][imad], atol=1e-7)

    assert np.isclose(twb2xt['mux'][ixt], twb2mad['mux'][imad], atol=1e-7)
    assert np.isclose(twb2xt['muy'][ixt], twb2mad['muy'][imad], atol=1e-7)

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_allclose

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()


@pytest.fixture(scope='module')
def line0(temp_context_default_mod):
    # The fixture `temp_context_default_mod` is defined in conftest.py and is
    # needed here as the non-empty context after the tests that invoke
    # this fixture would trigger an error (due to the empty context check).
    # See `conftest.py` for more details.
    with open(test_data_folder.joinpath(
            'hllhc15_noerrors_nobb/line_and_particle.json')) as f:
        dct = json.load(f)

    line0 = xt.Line.from_dict(dct['line'])
    line0.particle_ref = xp.Particles.from_dict(dct['particle'])
    line0.build_tracker()
    return line0


@pytest.fixture(scope='module')
def particles0(line0, temp_context_default_mod):
    # See the comment in `line0` fixture.
    particles0 = xp.generate_matched_gaussian_bunch(line=line0,
                                                    num_particles=50,
                                                    nemitt_x=2.5e-6,
                                                    nemitt_y=2.5e-6,
                                                    sigma_z=9e-2)
    return particles0


@for_all_test_contexts
def test_constructor(test_context):
    elements = [
        #xt.ParticlesMonitor(_context=test_context, start_at_turn=0, stop_at_turn=10, num_particles=42),
        #xt.LastTurnsMonitor(_context=test_context, n_last_turns=5, num_particles=42),
        xt.BeamPositionMonitor(_context=test_context, stop_at_turn=10),
    ]

    # test to_dict / from_dict
    for ee in elements:
        dd = ee.to_dict()
        nee = ee.__class__.from_dict(dd, _context=test_context)
        # Check that the two objects are bitwise identical
        if not isinstance(test_context, xo.ContextCpu):
            ee.move(_context=xo.ContextCpu())
            nee.move(_context=xo.ContextCpu())
        assert (ee._xobject._buffer.buffer[ee._xobject._offset:ee._xobject._size]
                - nee._xobject._buffer.buffer[
                    nee._xobject._offset:nee._xobject._size]).sum() == 0


@for_all_test_contexts
def test_monitor(test_context, line0, particles0):
    num_particles = len(particles0.x)
    line = line0.copy(_context=test_context)
    line.build_tracker(_context=test_context)
    particles = particles0.copy(_context=test_context)

    # Test implicit monitor
    num_turns = 30
    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)
    mon = line.record_last_track
    assert np.all(mon.x.shape == np.array([50, 30]))
    assert np.all(mon.at_turn[3, :] == np.arange(0, num_turns))
    assert np.all(mon.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(mon.at_element[:, :] == 0)
    assert np.all(mon.pzeta[:, 0] == particles0.pzeta)

    # Test explicit monitor passed to track
    monitor = xt.ParticlesMonitor(_context=test_context,
                                  start_at_turn=5, stop_at_turn=15,
                                  num_particles=num_particles)
    particles = particles0.copy(_context=test_context)
    num_turns = 30
    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=monitor)
    assert np.all(monitor.x.shape == np.array([50, 10]))
    assert np.all(monitor.at_turn[3, :] == np.arange(5, 15))
    assert np.all(monitor.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(monitor.at_element[:, :] == 0)
    assert np.all(monitor.pzeta[:, 0] == mon.pzeta[:, 5]) #5 in mon because the 0th entry of monitor is the 5th turn (5th entry in mon)

    # Test to_dict/from_dict round trip
    dct = monitor.to_dict()
    monitor2 = xt.ParticlesMonitor.from_dict(dct, _context=test_context)
    assert np.all(monitor2.x.shape == np.array([50, 10]))
    assert np.all(monitor2.at_turn[3, :] == np.arange(5, 15))
    assert np.all(monitor2.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(monitor2.at_element[:, :] == 0)
    assert np.all(monitor2.pzeta[:, 0] == mon.pzeta[:, 5]) #5 in mon because the 0th entry of monitor is the 5th turn (5th entry in mon)

    # Test explicit monitor used in in stand-alone mode
    mon2 = xt.ParticlesMonitor(_context=test_context,
                               start_at_turn=5, stop_at_turn=15,
                               num_particles=num_particles)
    particles = particles0.copy(_context=test_context)
    num_turns = 30
    for ii in range(num_turns):
        mon2.track(particles)
        line.track(particles)
    assert np.all(mon2.x.shape == np.array([50, 10]))
    assert np.all(mon2.at_turn[3, :] == np.arange(5, 15))
    assert np.all(mon2.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(mon2.at_element[:, :] == 0)
    assert np.all(mon2.pzeta[:, 0] == mon.pzeta[:, 5]) #5 in mon because the 0th entry of monitor is the 5th turn (5th entry in mon)

    # Test monitors with multiple frames
    monitor_multiframe = xt.ParticlesMonitor(_context=test_context,
                                             start_at_turn=5, stop_at_turn=10,
                                             n_repetitions=3,
                                             repetition_period=20,
                                             num_particles=num_particles)
    particles = particles0.copy(_context=test_context)
    num_turns = 100
    line.track(particles,
                  num_turns=num_turns,
                  turn_by_turn_monitor=monitor_multiframe)
    assert np.all(monitor_multiframe.x.shape == np.array([3, 50, 5]))
    assert np.all(monitor_multiframe.at_turn[1, 3, :] == np.arange(25, 30))
    assert np.all(monitor_multiframe.particle_id[2, :, 3] == np.arange(0,
                                                             num_particles))
    assert np.all(monitor_multiframe.at_element[:, :, :] == 0)
    assert np.all(monitor_multiframe.pzeta[0, :, 0] == mon.pzeta[:, 5]) #5 in mon because the 0th entry of monitor is the 5th turn (5th entry in mon)

    # Test monitors installed in a line
    monitor_ip5 = xt.ParticlesMonitor(start_at_turn=5, stop_at_turn=15,
                                      num_particles=num_particles)
    monitor_ip8 = xt.ParticlesMonitor(start_at_turn=5, stop_at_turn=15,
                                      num_particles=num_particles)
    line_w_monitor = line0.copy()
    line_w_monitor.insert_element(index='ip5', element=monitor_ip5, name='mymon5')
    line_w_monitor.insert_element(index='ip8', element=monitor_ip8, name='mymon8')

    line_w_monitor.build_tracker(_context=test_context)

    particles = particles0.copy(_context=test_context)
    num_turns = 50
    line_w_monitor.track(particles, num_turns=num_turns)

    assert np.all(monitor_ip5.x.shape == np.array([50, 10]))
    assert np.all(monitor_ip5.at_turn[3, :] == np.arange(5, 15))
    assert np.all(monitor_ip5.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(monitor_ip5.at_element[:, :]
                        == line_w_monitor.element_names.index('ip5') - 1)

    assert np.all(monitor_ip8.x.shape == np.array([50, 10]))
    assert np.all(monitor_ip8.at_turn[3, :] == np.arange(5, 15))
    assert np.all(monitor_ip8.particle_id[:, 3] == np.arange(0, num_particles))
    assert np.all(monitor_ip8.at_element[:, :]
                        == line_w_monitor.element_names.index('ip8') - 1)


@for_all_test_contexts
def test_last_turns_monitor(test_context):
    particles = xp.Particles(p0c=6.5e12, x=[1,2,3,4,5,6], _context=test_context)

    monitor = xt.LastTurnsMonitor(n_last_turns=5, particle_id_range=(1, 5), _context=test_context)

    line = xt.Line([monitor])
    line.build_tracker(_context=test_context)

    for turn in range(10):

        line.track(particles, num_turns=1)

        # Note that indicees are re-ordered upon particle loss on CPU contexts,
        # so sort before manipulation
        if isinstance(test_context, xo.ContextCpu):
            particles.sort(interleave_lost_particles=True)

        particles.x[0] += 1
        particles.x[1] -= 1
        particles.x[2] += 2
        particles.x[3] -= 2
        particles.x[4] += 3
        particles.x[5] -= 3
        if turn == 2:
            particles.state[1] = 0
        if turn == 4:
            particles.state[2] = 0
        if turn == 6:
            particles.state[3] = 0

        if isinstance(test_context, xo.ContextCpu):
            particles.reorganize()


    assert np.all(monitor.particle_id == np.array([[0,0,1,1,1],[2]*5,[3]*5,[4]*5]))
    assert np.all(monitor.at_turn == np.array([np.clip(n-np.arange(4,-1,-1),0,None) for n in (2,4,6,9)]))
    assert np.all(monitor.x == np.array([[0,0,2,1,0],[3,5,7,9,11],[0,-2,-4,-6,-8],[20,23,26,29,32]]))


@for_all_test_contexts
def test_beam_profile_monitor(test_context):
    gen = np.random.default_rng(seed=38715345)

    npart = 512 # must be even and >= 512
    x = gen.normal(size=npart+4) # generate a few more to test "num_particles"
    y = gen.normal(size=npart+4)

    particles = xp.Particles(
        p0c=6.5e12,
        x=x,
        y=y,
        zeta=-2.99792458e8*np.tile([0.0, 0.5], x.size//2),
        _context=test_context,
    )

    nbins = 100

    monitor = xt.BeamProfileMonitor(
        num_particles=npart,
        start_at_turn=0,
        stop_at_turn=10,
        frev=1,
        sampling_frequency=2,
        n=nbins,
        x_range=5,
        y_range=(-4,2),
        _context=test_context,
    )

    line = xt.Line([monitor])
    line.build_tracker(_context=test_context)

    for turn in range(11): # track a few more than we record to test "stop_at_turn"
        # Note that indicees are re-ordered upon particle loss on CPU contexts,
        # so sort before manipulation
        if isinstance(test_context, xo.ContextCpu):
            particles.sort(interleave_lost_particles=True)

        # manipulate particles for testing
        particles.x[0] += 0.1
        particles.y[0] -= 0.1
        if turn == 8:
            particles.state[256:] = 0
        if turn == 9:
            particles.state[:] = 0

        if isinstance(test_context, xo.ContextCpu):
            particles.reorganize()

        # track and monitor
        line.track(particles, num_turns=1)


    # Check against expected values
    expected_x_intensity = np.zeros((20, nbins))
    expected_y_intensity = np.zeros((20, nbins))
    for turn in range(10):
        for sub in range(2):
            lim = {8:256, 9:0}.get(turn, npart) # consider dead particles in last turns
            x_sub = np.copy(x[:lim][sub::2])
            y_sub = np.copy(y[:lim][sub::2])
            if sub == 0 and lim > 0:
                x_sub[0] += 0.1 * (turn+1)
                y_sub[0] -= 0.1 * (turn+1)
            # benchmark against numpy's histogram function
            hist_x, edges_x = np.histogram(x_sub, bins=nbins, range=(-2.5, 2.5))
            hist_y, edges_y = np.histogram(y_sub, bins=nbins, range=(-4, 2))
            expected_x_intensity[2*turn+sub, :] = hist_x
            expected_y_intensity[2*turn+sub, :] = hist_y

    assert_allclose(monitor.x_edges, edges_x, err_msg="Monitor x_edges does not match expected values")
    assert_allclose(monitor.y_edges, edges_y, err_msg="Monitor y_edges does not match expected values")    
    assert_allclose(monitor.x_grid, (edges_x[1:]+edges_x[:-1])/2, err_msg="Monitor x_grid does not match expected values")
    assert_allclose(monitor.y_grid, (edges_y[1:]+edges_y[:-1])/2, err_msg="Monitor y_grid does not match expected values")
    assert_allclose(monitor.x_intensity, expected_x_intensity, err_msg="Monitor x_intensity does not match expected values")
    assert_allclose(monitor.y_intensity, expected_y_intensity, err_msg="Monitor y_intensity does not match expected values")


@for_all_test_contexts
def test_collective_ebe_monitor(test_context, line0, particles0):
    num_turns = 30

    # Turn-by-turn mode
    monitor_mode = True
    # Line without collective elements
    line = line0.copy()
    line.build_tracker(_context=test_context)
    particles = particles0.copy(_context=test_context)

    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=monitor_mode)
    recoded_track_x = line.record_last_track.x

    # Line with collective elements
    line_collective = line0.copy()
    line_collective.elements[2].iscollective = True # Thick element (Drift)
    line_collective.elements[2].move(_context=test_context)
    line_collective.elements[12000].iscollective = True
    line_collective.elements[12000].move(_context=test_context)
    line_collective.build_tracker(_context=test_context)
    particles_collective = particles0.copy(_context=test_context)

    line_collective.track(
        particles_collective,
        num_turns=num_turns,
        turn_by_turn_monitor=monitor_mode,
    )
    recoded_track_x_collective = line_collective.record_last_track.x

    xo.assert_allclose(recoded_track_x, recoded_track_x_collective)

    # Element by element mode
    monitor_mode = 'ONE_TURN_EBE'
    # Line without collective elements
    line = line0.copy()
    line.build_tracker(_context=test_context)
    particles = particles0.copy(_context=test_context)

    line.track(particles, num_turns=num_turns, turn_by_turn_monitor=monitor_mode)
    recoded_track_x = line.record_last_track.x

    # Line with collective elements
    line_collective = line0.copy()
    line_collective.elements[2].iscollective = True # Thick element (Drift)
    line_collective.elements[2].move(_context=test_context)
    line_collective.elements[12000].iscollective = True
    line_collective.elements[12000].move(_context=test_context)
    line_collective.build_tracker(_context=test_context)
    particles_collective = particles0.copy(_context=test_context)

    line_collective.track(
        particles_collective,
        num_turns=num_turns,
        turn_by_turn_monitor=monitor_mode,
    )
    recoded_track_x_collective = line_collective.record_last_track.x
    xo.assert_allclose(recoded_track_x, recoded_track_x_collective)


@for_all_test_contexts
def test_beam_size_monitor(test_context):

    npart = 512 # must be even and >= 512
    x = np.linspace(-0.1, 0.1, npart+4)**2 # generate a few more than we record to test "num_particles"

    particles = xp.Particles(
        p0c=6.5e12,
        x=x,
        y=10*x,
        zeta=-2.99792458e8*np.tile([0.0, 0.5], (npart+4)//2),
        _context=test_context,
    )

    monitor = xt.BeamSizeMonitor(
        num_particles=npart,
        start_at_turn=0,
        stop_at_turn=10,
        frev=1,
        sampling_frequency=2,
        _context=test_context,
    )

    line = xt.Line([monitor])
    line.build_tracker(_context=test_context)

    for turn in range(11): # track a few more than we record to test "stop_at_turn"
        # Note that indicees are re-ordered upon particle loss on CPU contexts,
        # so sort before manipulation
        if isinstance(test_context, xo.ContextCpu):
            particles.sort(interleave_lost_particles=True)

        # manipulate particles for testing
        particles.x[0] += 0.005
        particles.y[0] += 0.05
        if turn == 8:
            particles.state[256:] = 0
        if turn == 9:
            particles.state[:] = 0

        if isinstance(test_context, xo.ContextCpu):
            particles.reorganize()

        # track and monitor
        line.track(particles, num_turns=1)

    # Check against expected values
    expected_count = np.tile([npart//2, npart//2], 10)
    expected_count[16:18] = 128
    expected_count[18:20] = 0
    assert_equal(monitor.count, expected_count, err_msg="Monitor count does not match expected particle count")

    x = x[:npart].reshape((-1, 2))
    expected_x_sum = np.tile(np.sum(x, axis=0), 10)
    expected_x_sum[0::2] += 0.005*np.arange(1, 11)
    expected_x_sum[16:18] -= np.sum(x[128:, :], axis=0)
    expected_x_sum[18:20] = 0
    assert_allclose(monitor.x_sum, expected_x_sum, err_msg="Monitor x sum does not match expected values")
    assert_allclose(monitor.y_sum, 10*expected_x_sum, err_msg="Monitor y sum does not match expected values")

    expected_x_mean = np.zeros(20)
    expected_x_mean[:18] = expected_x_sum[:18]/expected_count[:18]
    expected_x_mean[18:20] = np.nan
    assert_allclose(monitor.x_mean, expected_x_mean, err_msg="Monitor x mean does not match expected values")
    assert_allclose(monitor.y_mean, 10*expected_x_mean, err_msg="Monitor y mean does not match expected values")

    expected_x2_sum = np.tile(np.sum(x**2, axis=0), 10)
    expected_x2_sum[0::2] += 2*x[0,0]*0.005*np.arange(1, 11) + (0.005*np.arange(1, 11))**2
    expected_x2_sum[16:18] -= np.sum(x[128:, :]**2, axis=0)
    expected_x2_sum[18:20] = 0
    assert_allclose(monitor.x2_sum, expected_x2_sum, err_msg="Monitor x^2 sum does not match expected values")
    assert_allclose(monitor.y2_sum, 100*expected_x2_sum, err_msg="Monitor y^2 sum does not match expected values")

    expected_x2_mean = np.zeros(20)
    expected_x2_mean[:18] = expected_x2_sum[:18]/expected_count[:18]
    expected_x2_mean[18:20] = np.nan
    expected_x_var = expected_x2_mean - expected_x_mean**2  # Var(x) = mean(x**2) - mean(x)**2
    assert_allclose(monitor.x_var, expected_x_var, err_msg="Monitor x variance does not match expected values")
    assert_allclose(monitor.y_var, 100*expected_x_var, err_msg="Monitor y variance does not match expected values")

    expected_x_std = expected_x_var**0.5
    assert_allclose(monitor.x_std, expected_x_std, err_msg="Monitor x standard deviation does not match expected values")
    assert_allclose(monitor.y_std, 10*expected_x_std, err_msg="Monitor y standard deviation does not match expected values")


@for_all_test_contexts
def test_beam_position_monitor(test_context):

    npart = 512 # must be even and >= 512
    particles = xp.Particles(
        p0c=6.5e12,
        x=0.1*np.arange(npart+4), # generate a few more than we record to test "num_particles"
        y=-0.1*np.arange(npart+4),
        zeta=-2.99792458e8*np.tile([0.0, 0.5], (npart+4)//2),
        _context=test_context,
    )

    monitor = xt.BeamPositionMonitor(
        num_particles=npart,
        start_at_turn=0,
        stop_at_turn=10,
        frev=1,
        sampling_frequency=2,
        _context=test_context,
    )

    line = xt.Line([monitor])
    line.build_tracker(_context=test_context)

    for turn in range(11): # track a few more than we record to test "stop_at_turn"
        # Note that indicees are re-ordered upon particle loss on CPU contexts,
        # so sort before manipulation
        if isinstance(test_context, xo.ContextCpu):
            particles.sort(interleave_lost_particles=True)

        # manipulate particles for testing
        particles.x[0] += 0.5
        particles.y[0] -= 0.5
        if turn == 8:
            particles.state[256:] = 0
        if turn == 9:
            particles.state[:] = 0

        if isinstance(test_context, xo.ContextCpu):
            particles.reorganize()

        # track and monitor
        line.track(particles, num_turns=1)

    expected_count = np.tile([npart//2, npart//2], 10)
    expected_count[16:18] = 128
    expected_count[18:20] = 0
    assert_equal(monitor.count, expected_count, err_msg="Monitor count does not match expected particle count")

    expected_x_sum = np.tile([0.1*(npart-2)*npart/4, 0.1*npart*npart/4], 10)
    expected_x_sum[0::2] += 0.5*np.arange(1, 11)
    expected_x_sum[16:18] = [1625.6 + 4.5, 1638.4]
    expected_x_sum[18:20] = 0
    assert_allclose(monitor.x_sum, expected_x_sum, err_msg="Monitor x sum does not match expected values")
    assert_allclose(monitor.y_sum, -expected_x_sum, err_msg="Monitor y sum does not match expected values")

    expected_x_centroid = np.zeros(20)
    expected_x_centroid[:18] = expected_x_sum[:18]/expected_count[:18]
    expected_x_centroid[18:20] = np.nan
    assert_allclose(monitor.x_cen, expected_x_centroid, err_msg="Monitor x centroid does not match expected values")
    assert_allclose(monitor.y_cen, -expected_x_centroid, err_msg="Monitor y centroid does not match expected values")



import numpy as np
import xtrack as xt

bend = xt.Bend(k0=0.4, h=0.3, length=1)

elements = {
    # correct case
    'a0': bend,
    'a1': xt.Replica(parent_name='a0'),
    'a2': xt.Replica(parent_name='a1'),
    # create a simple loop
    'b0': xt.Replica(parent_name='b0'),
    # loop with two replicas
    'c0': xt.Replica(parent_name='c1'),
    'c1': xt.Replica(parent_name='c0'),
    # bigger loop
    'd0': xt.Replica(parent_name='d1'),
    'd1': xt.Replica(parent_name='d2'),
    'd2': xt.Replica(parent_name='d0'),
    'd3': xt.Replica(parent_name='d1'),
}

line = xt.Line(elements=elements, element_names=list(elements.keys()))

ok = ['a1', 'a2']
for kk in ok:
    assert line[kk].resolve(line) is line['a0']

error = ['b0', 'c0', 'c1', 'd0', 'd1', 'd2', 'd3']
for kk in error:
    try:
        line[kk].resolve(line)
    except RecursionError:
        pass
    else:
        raise Exception(f'Element {kk} should not be resolvable')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import sixtracktools
import xpart as xp
import xobjects as xo
import xtrack as xt

##################
# Get a sequence #
##################

six = sixtracktools.SixInput(".")
line = six.generate_xtrack_line()


######################
# Get some particles #
######################
sixdump = sixtracktools.SixDump101("res/dump3.dat")
part0= xp.Particles.from_dict(sixdump[0::2][0].get_minimal_beam())

# Force active state
part0.state = 1

with open('line_and_particle.json', 'w') as fid:
    json.dump({
        'line': line.to_dict(),
        'particle': part0.to_dict()},
        fid, cls=xo.JEncoder,
        indent=4)




import numpy as np

import xtrack as xt

from . import lumi


class TargetLuminosity(xt.Target):

    def __init__(self, ip_name, luminosity, tol, num_colliding_bunches,
                 num_particles_per_bunch, nemitt_x, nemitt_y, sigma_z, f_rev,
                 crab=None, weight=None, log=True):

        if log:
            value = np.log10(luminosity)
            tol = np.log10(tol/luminosity + 1) # equivalent (luminosity - target)/target < tol / target
        else:
            value = luminosity
            tol = tol

        xt.Target.__init__(self, self.compute_luminosity, value, tol=tol)

        self.ip_name = ip_name
        self.num_colliding_bunches = num_colliding_bunches
        self.num_particles_per_bunch = num_particles_per_bunch
        self.nemitt_x = nemitt_x
        self.nemitt_y = nemitt_y
        self.sigma_z = sigma_z
        self.f_rev = f_rev
        self.crab = crab
        self.weight = weight
        self.log = log

    def __repr__(self):
        return f'TargetLuminosity(ip_name={self.ip_name}, luminosity={self.value}, tol={self.tol})'

    @property
    def weight(self):
        if self._weight is None:
            return 1 / self.value
        else:
            return self._weight

    @weight.setter
    def weight(self, value):
        self._weight = value

    def compute_luminosity(self, tw):
        assert len(tw._line_names) == 2
        out = lumi.luminosity_from_twiss(
            n_colliding_bunches=self.num_colliding_bunches,
            num_particles_per_bunch=self.num_particles_per_bunch,
            ip_name=self.ip_name,
            nemitt_x=self.nemitt_x,
            nemitt_y=self.nemitt_y,
            sigma_z=self.sigma_z,
            twiss_b1=tw[tw._line_names[0]],
            twiss_b2=tw[tw._line_names[1]],
            f_rev=self.f_rev,
            crab=self.crab)
        if self.log:
            out = np.log10(out)
        return out

class TargetSeparationOrthogonalToCrossing(xt.Target):

    def __init__(self, ip_name):
        xt.Target.__init__(self, tar=self.projection, value=0, tol=1e-6, scale=1)
        self.ip_name = ip_name

    def __repr__(self):
        return f'TargetSeparationOrthogonalToCrossing(ip_name={self.ip_name})'

    def projection(self, tw):
        assert len(tw._line_names) == 2
        ip_name = self.ip_name

        twb1 = tw[tw._line_names[0]]
        twb2 = tw[tw._line_names[1]].reverse()

        diff_px = twb1['px', ip_name] - twb2['px', ip_name]
        diff_py = twb1['py', ip_name] - twb2['py', ip_name]

        diff_p_mod = np.sqrt(diff_px**2 + diff_py**2)

        diff_x = twb1['x', ip_name] - twb2['x', ip_name]
        diff_y = twb1['y', ip_name] - twb2['y', ip_name]

        diff_r_mod = np.sqrt(diff_x**2 + diff_y**2)

        return (diff_x*diff_px + diff_y*diff_py)/(diff_p_mod*diff_r_mod)

class TargetSeparation(xt.Target):

    def __init__(self, ip_name, separation=None, separation_norm=None,
                 plane=None, nemitt_x=None, nemitt_y=None, tol=None, scale=None,
                 ineq_sign=None):


        # For now nemitt is a scalar, we can move to a tuple for different
        # emittances for the two beams

        if separation is None and separation_norm is None:
            raise ValueError('Either separation or separation_norm must be '
                             'provided')

        if separation is not None:
            value = separation
        elif separation_norm is not None:
            value = separation_norm

        if separation_norm is not None and (nemitt_x is None and nemitt_y is None):
            raise ValueError('nemitt must be provided if separation_norm is '
                             'provided')

        if plane is None: # In the future plane=None could be used to set the distance between 
                          # the two beams computed as sqrt(dx**2 + dy**2)
            raise ValueError('plane must be provided')
        assert plane in ['x', 'y']
        assert ineq_sign in [None, '<', '>']

        xt.Target.__init__(self, tar=self.get_separation, value=value, tol=tol,
                            scale=scale)

        self.ip_name = ip_name
        self.separation = separation
        self.separation_norm = separation_norm
        self.plane = plane
        self.nemitt_x = nemitt_x
        self.nemitt_y = nemitt_y
        self.ineq_sign = ineq_sign

    def __repr__(self):
        return f'TargetSeparation(ip_name={self.ip_name}, value={self.value}, tol={self.tol}, ineq_sign={self.ineq_sign})'

    def get_separation(self, tw):

        assert len(tw._line_names) == 2
        tw1 = tw[tw._line_names[0]]
        tw2 = tw[tw._line_names[1]].reverse()

        if self.separation is not None:
            out = np.abs(tw1[self.plane, self.ip_name] - tw2[self.plane, self.ip_name])
        elif self.separation_norm is not None:
            nemitt = self.nemitt_x if self.plane == 'x' else self.nemitt_y
            sigma1 = np.sqrt(
                tw1['bet'+self.plane, self.ip_name] * nemitt
                / tw1.particle_on_co.gamma0[0] / tw1.particle_on_co.beta0[0])
            sigma2 = np.sqrt(
                tw2['bet'+self.plane, self.ip_name] * nemitt
                / tw2.particle_on_co.gamma0[0] / tw2.particle_on_co.beta0[0])

            sigma = np.sqrt(sigma1*sigma2) # geometric mean of sigmas
            sep_norm = np.abs((tw1[self.plane, self.ip_name] - tw2[self.plane, self.ip_name]
                        )) / sigma

            out = sep_norm

        if self.ineq_sign is not None:
            raise NotImplementedError # untested
            if self.ineq_sign == '<':
                if out < self.value:
                    out = self.value
            elif self.ineq_sign == '>':
                if out > self.value:
                    out = self.value

        return out


import numpy as np
import xtrack as xt
import xobjects as xo
import xpart as xp

import RF_Track as RFT

# Bend parameters
clight = 299792458 # m/s
p0c = 1.2e9  # eV
lbend = 3 # m
angle = np.pi / 2 # rad
rho = lbend / angle # m
By = p0c / rho / clight # T

#############################################
#######  RF-Track's part starts here  #######
#############################################

# Define the RF-Track element
sbend = RFT.SBend(lbend, angle, p0c / 1e6)

#############################################
#######  RF-Track's part ends here    #######
#############################################

# Back to Xsuite
pi = np.pi
lbend = 3
elements = {
    'd1.1':  xt.Drift(length=1),
    'mb1.1': xt.RFT_Element(element=sbend),
    'd2.1':  xt.Drift(length=1),

    'mqd.1': xt.Quadrupole(length=0.3, k1=-0.7),
    'd3.1':  xt.Drift(length=1),
    'mb2.1': xt.RFT_Element(element=sbend),
    'd4.1':  xt.Drift(length=1),

    'd1.2':  xt.Drift(length=1),
    'mb1.2': xt.RFT_Element(element=sbend),
    'd2.2':  xt.Drift(length=1),

    'mqd.2': xt.Quadrupole(length=0.3, k1=-0.7),
    'd3.2':  xt.Drift(length=1),
    'mb2.2': xt.RFT_Element(element=sbend),
    'd4.2':  xt.Drift(length=1),
}

# Build the ring
line = xt.Line(elements=elements, element_names=list(elements.keys()))
line.particle_ref = xt.Particles(p0c=p0c, mass0=xt.PROTON_MASS_EV)
line.configure_bend_model(core='full', edge=None)

# track
## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

## Build particle object on context
n_part = 200

rng = np.random.default_rng(2021)

particles = xp.Particles(p0c=p0c, #eV
                        q0=1, mass0=xp.PROTON_MASS_EV,
                        x=rng.uniform(-1e-3, 1e-3, n_part),
                        px=rng.uniform(-1e-5, 1e-5, n_part),
                        y=rng.uniform(-2e-3, 2e-3, n_part),
                        py=rng.uniform(-3e-5, 3e-5, n_part),
                        zeta=rng.uniform(-1e-2, 1e-2, n_part),
                        delta=rng.uniform(-1e-3, 1e-3, n_part),
                        _context=context)

print('tracking starts')
line.track(particles)
print('tracking ends')

from scipy.io import savemat
s = np.transpose(np.array([particles.x, particles.px, particles.y, particles.py, particles.zeta, particles.delta]))
dict = { "s": s }
savemat("particles_rft_sbend.mat", dict)



import xtrack as xt
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9, mass0=xt.PROTON_MASS_EV)
env['pi'] = np.pi

env['l_bend'] = 3.5
env['l_quad'] = 1.
env['l_cell'] = 20.
env['n_bends'] = 24.

env['h_bend']= 'pi / n_bends / l_bend'

env.new('mq', xt.Quadrupole, length='l_quad')
env.new('mb', xt.Bend, length='l_bend', h='h_bend', k0='h_bend')

env.new('mqf', 'mq', k1=0.1)
env.new('mqd', 'mq', k1=-0.1)

# We create a line which defines half of a FODO cell
arc_half_cell = env.new_line(components=[
    env.place('mqf'),
    env.place('mb', at='l_cell/4 - (l_bend/2 + 0.2)', from_='mqf'),
    env.place('mb', at='l_cell/4 + (l_bend/2 + 0.2)', from_='mqf'),
    env.place('mqd', at='l_cell/2 - l_quad/2', from_='mqf'),
    ])
arc_half_cell.survey().plot() # plots the layout of the created line
arc_half_cell.get_table()
# is:
#
# name                   s element_type isthick isreplica parent_name ...
# mqf                    0 Quadrupole      True     False None
# drift_1                1 Drift           True     False None
# mb::0                1.8 Bend            True     False None
# drift_2              5.3 Drift           True     False None
# mb::1                5.7 Bend            True     False None
# drift_3              9.2 Drift           True     False None
# mqd                  9.5 Quadrupole      True     False None
# _end_point          10.5                False     False None

# We can mirror a line by simply using the minus sign
(-arc_half_cell).get_table()
# is:
# name                   s element_type isthick isreplica parent_name ...
# mqd                    0 Quadrupole      True     False None
# drift_3                1 Drift           True     False None
# mb::0                1.3 Bend            True     False None
# drift_2              4.8 Drift           True     False None
# mb::1                5.2 Bend            True     False None
# drift_1              8.7 Drift           True     False None
# mqf                  9.5 Quadrupole      True     False None
# _end_point          10.5                False     False None

# We concatenate the two to obtain a full cell
arc_cell = -arc_half_cell + arc_half_cell
arc_cell.get_table()
# gives the following (note that the occurrence of each repeated element is
# indicated together with the element name):
#
# Table: 15 rows, 8 cols
# name                   s element_type isthick isreplica parent_name ...
# mqd::0                 0 Quadrupole      True     False None
# drift_3::0             1 Drift           True     False None
# mb::0                1.3 Bend            True     False None
# drift_2::0           4.8 Drift           True     False None
# mb::1                5.2 Bend            True     False None
# drift_1::0           8.7 Drift           True     False None
# mqf::0               9.5 Quadrupole      True     False None
# mqf::1              10.5 Quadrupole      True     False None
# drift_1::1          11.5 Drift           True     False None
# mb::2               12.3 Bend            True     False None
# drift_2::1          15.8 Drift           True     False None
# mb::3               16.2 Bend            True     False None
# drift_3::1          19.7 Drift           True     False None
# mqd::1                20 Quadrupole      True     False None
# _end_point            21                False     False None

# We create another line defining a single straight cell
straight_cell = env.new_line(components=[
    env.new('s.scell', xt.Marker), # At start cell
    env.place('mqf'),
    env.place('mqd', at='l_cell/2', from_='mqf'), # At mid cell
    env.new('e.scell', xt.Marker, at='l_cell')])

# We a ring composing arcs and straight sections
ring = 6 * (2 * arc_cell + 3 * straight_cell)

# We plot the layout of the ring
ring.survey().plot()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt


line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(
                    mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()
line.vars['on_disp'] = 1

tw = line.twiss()

init_at = 'e.cell.45.b1'

x = tw['x', init_at]
y = tw['y', init_at]
px = tw['px', init_at]
py = tw['py', init_at]
zeta = tw['zeta', init_at]
delta = tw['delta', init_at]
betx = tw['betx', init_at]
bety = tw['bety', init_at]
alfx = tw['alfx', init_at]
alfy = tw['alfy', init_at]
dx = tw['dx', init_at]
dy = tw['dy', init_at]
dpx = tw['dpx', init_at]
dpy = tw['dpy', init_at]
mux = tw['mux', init_at]
muy = tw['muy', init_at]
muzeta = tw['muzeta', init_at]
dzeta = tw['dzeta', init_at]
bets = tw.bets0
reference_frame = 'proper'

tw_init = xt.TwissInit(element_name=init_at,
    x=x, px=px, y=y, py=py, zeta=zeta, delta=delta,
    betx=betx, bety=bety, alfx=alfx, alfy=alfy,
    dx=dx, dy=dy, dpx=dpx, dpy=dpy,
    mux=mux, muy=muy, muzeta=muzeta, dzeta=dzeta,
    bets=bets, reference_frame=reference_frame)
tw_test = line.twiss(start=init_at, end='ip6', init=tw_init)


import matplotlib.pyplot as plt

fig1 = plt.figure(1, figsize=(6.4, 4.8*1.5))
spbet = plt.subplot(3,1,1)
spco = plt.subplot(3,1,2, sharex=spbet)
spdisp = plt.subplot(3,1,3, sharex=spbet)

spbet.plot(tw['s'], tw['betx'])
spbet.plot(tw['s'], tw['bety'])
spbet.plot(tw_test['s'], tw_test['betx'], '--')
spbet.plot(tw_test['s'], tw_test['bety'], '--')

spco.plot(tw['s'], tw['x'])
spco.plot(tw['s'], tw['y'])
spco.plot(tw_test['s'], tw_test['x'], '--')
spco.plot(tw_test['s'], tw_test['y'], '--')

spdisp.plot(tw['s'], tw['dx'])
spdisp.plot(tw['s'], tw['dy'])
spdisp.plot(tw_test['s'], tw_test['dx'], '--')
spdisp.plot(tw_test['s'], tw_test['dy'], '--')

spbet.set_ylabel(r'$\beta_{x,y}$ [m]')
spco.set_ylabel(r'(Closed orbit)$_{x,y}$ [m]')
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

assert tw_test.name[-1] == '_end_point'
tw_part = tw.rows['e.cell.45.b1':'ip6']

tw_test = tw_test.rows[:-1]
assert np.all(tw_test.name == tw_part.name)

atols = dict(
    alfx=1e-8, alfy=1e-8,
    dzeta=1e-3, dx=1e-4, dy=1e-4, dpx=1e-5, dpy=1e-5,
    nuzeta=1e-5, dx_zeta=1e-4, dy_zeta=1e-4, betx2=1e-3, bety1=1e-3,
    muzeta=1e-7,
)

rtols = dict(
    alfx=5e-9, alfy=5e-8,
    betx=1e-8, bety=1e-8, betx1=1e-8, bety2=1e-8,
    gamx=1e-8, gamy=1e-8,
)

atol_default = 1e-11
rtol_default = 1e-9


for kk in tw_test._data.keys():
    if kk in ['name', 'W_matrix', 'particle_on_co', 'values_at', 'method',
            'radiation_method', 'reference_frame', 'orientation']:
        continue # tested separately
    atol = atols.get(kk, atol_default)
    rtol = rtols.get(kk, rtol_default)
    assert np.allclose(
        tw_test._data[kk], tw_part._data[kk], rtol=rtol, atol=atol)

assert tw_test.values_at == tw_part.values_at == 'entry'
assert tw_test.radiation_method == tw_part.radiation_method == 'full'
assert tw_test.reference_frame == tw_part.reference_frame == 'proper'

W_matrix_part = tw_part.W_matrix
W_matrix_test = tw_test.W_matrix

for ss in range(W_matrix_part.shape[0]):
    this_part = W_matrix_part[ss, :, :]
    this_test = W_matrix_test[ss, :, :]

    for ii in range(4):
        assert np.isclose((np.linalg.norm(this_part[ii, :] - this_test[ii, :])
                        /np.linalg.norm(this_part[ii, :])), 0, atol=3e-4)

plt.show()




import xtrack as xt
import xobjects as xo

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
env.vars.default_to_zero = True

env.new('mq', 'Quadrupole', length=0.8)
env.new('bpm', 'Marker')
env.new('corrector', 'Multipole', knl=[0])

line = env.new_line(components=[

    env.new('line.start', 'Marker'),
    env.new('line.end', 'Marker', at=12.),

    env.new('mq1', 'mq', k1=0.2, at=3.),
    env.new('mq2', 'mq', k1=-0.2, at=5.),
    env.new('mq3', 'mq', k1=0.2, at=7.),
    env.new('mq4', 'mq', k1=-0.2, at=9.),

    env.new('bpm.s.1', 'bpm', at=0.1),
    env.new('bpm.s.2', 'bpm', at=0.5),
    env.new('bpm.q.1', 'bpm', at='mq1@start'),
    env.new('bpm.q.2', 'bpm', at='mq2@start'),
    env.new('bpm.q.3', 'bpm', at='mq3@start'),
    env.new('bpm.q.4', 'bpm', at='mq4@start'),
    env.new('bpm.e.1', 'bpm', at=11.5),
    env.new('bpm.e.2', 'bpm', at=11.9),

    env.new('corr1', 'corrector', at=1., knl=['k0l_corr1'], ksl=['k0sl_corr1']),
    env.new('corr2', 'corrector', at=2., knl=['k0l_corr2'], ksl=['k0sl_corr2']),
    env.new('corr3', 'corrector', at=10., knl=['k0l_corr3'], ksl=['k0sl_corr3']),
    env.new('corr4', 'corrector', at=11., knl=['k0l_corr4'], ksl=['k0sl_corr4']),
])

# Define monitors and correctors for orbit steering
line.steering_monitors_x = ['bpm.s.1', 'bpm.s.2',
                            'bpm.q.1', 'bpm.q.2', 'bpm.q.3', 'bpm.q.4',
                            'bpm.e.1', 'bpm.e.2']
line.steering_correctors_x = ['corr1', 'corr2', 'corr3', 'corr4']
line.steering_monitors_y = line.steering_monitors_x
line.steering_correctors_y = line.steering_correctors_x

# Twiss without misalignments
tw0 = line.twiss4d()

# Misalign all quadrupoles
env.set(['mq1', 'mq2', 'mq3', 'mq4'], shift_x=1e-3, shift_y=2e-3, rot_s_rad=0.1)

# Define BPM alignment
bpm_alignment ={
    'bpm.q.1': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
    'bpm.q.2': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
    'bpm.q.3': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
    'bpm.q.4': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
}

# Correct orbit taking into account BPM alignment (centers the beam in all quadrupoles)
correction = line.correct_trajectory(twiss_table=tw0,
                                     monitor_alignment=bpm_alignment, # <--BPM alignment
                                     run=False)
correction.correct()

#!end-doc-part

tw_corr = line.twiss4d()

xo.assert_allclose(correction.x_correction.shift_x_monitors,
                   [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.x_correction.shift_y_monitors,
                   [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.y_correction.shift_x_monitors,
                   [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.y_correction.shift_y_monitors,
                   [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)

# Data from previous step can be found in:
correction.correct() # Some more steps to log the position
xo.assert_allclose(correction.x_correction._position_before,0, rtol=0, atol=1e-10)
xo.assert_allclose(correction.y_correction._position_before,0, rtol=0, atol=1e-10)

correction.clear_correction_knobs()

for nn in ['corr1', 'corr2', 'corr3', 'corr4']:
    assert line[nn].knl[0] == 0

correction.thread(ds_thread = 10)
tw_thread = line.twiss4d()

xo.assert_allclose(correction.x_correction.shift_x_monitors,
                   [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.x_correction.shift_y_monitors,
                   [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.y_correction.shift_x_monitors,
                   [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
xo.assert_allclose(correction.y_correction.shift_y_monitors,
                   [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)

# Data from previous step can be found in:
correction.correct() # Some more steps to log the position
xo.assert_allclose(correction.x_correction._position_before,0, rtol=0, atol=1e-9)
xo.assert_allclose(correction.y_correction._position_before,0, rtol=0, atol=1e-9)

xo.assert_allclose(tw_corr.rows['mq1':'corr3'].x, 1e-3, rtol=0, atol=1e-9)
xo.assert_allclose(tw_corr.rows['mq1':'corr3'].y, 2e-3, rtol=0, atol=1e-9)
xo.assert_allclose(tw_thread.rows['mq1':'corr3'].x, 1e-3, rtol=0, atol=1e-4)
xo.assert_allclose(tw_thread.rows['mq1':'corr3'].y, 2e-3, rtol=0, atol=1e-4)

xo.assert_allclose(tw_corr.x[0], 0, rtol=0, atol=1e-9)
xo.assert_allclose(tw_corr.y[0], 0, rtol=0, atol=1e-9)
xo.assert_allclose(tw_thread.x[0], 0, rtol=0, atol=1e-4)
xo.assert_allclose(tw_thread.y[0], 0, rtol=0, atol=1e-4)

import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1)
tw_corr.plot('x y', figure=fig1, grid=False)
plt.suptitle('After correction')

fig2 = plt.figure(2)
tw_thread.plot('x y', figure=fig2, grid=False)
plt.suptitle('After threading')

plt.show()




# RadiationIntegral Class
# Works for both thick and thin elements
# By Simon Fanica Buijsman
# Date: 2024-11-28
# simon.fanica.buijsman@cern.ch


# TODO: Check if the curvature is calculated correctly for off-momentum particles.
# The current challenge is to see if this simulation is valid for particles that do not follow the design orbit.
# So far, I've only tested the simulation for the design trajectory. In this case, the curvature is trivial.


#### ---------------------------------------------------------------------------------------------------------------------------------
#### ---------------------------------------------------------------------------------------------------------------------------------
import numpy as np

# Physical constants required to calculate the physical quantities associated with the radiation integrals.
from scipy.constants import c as clight, hbar, electron_volt


#dict_keys(['name', 's', 'x', 'px', 'y', 'py', 'zeta', 'delta', 'ptau', 'W_matrix', 'kin_px', 'kin_py', 'kin_ps', 'kin_xprime', 'kin_yprime',
#           'betx', 'bety', 'alfx', 'alfy', 'gamx', 'gamy', 'dx', 'dpx', 'dy', 'dpy', 'dx_zeta', 'dpx_zeta', 'dy_zeta', 'dpy_zeta', 'betx1', 
#           'bety1', 'betx2', 'bety2', 'mux', 'muy', 'muzeta', 'nux', 'nuy', 'nuzeta', 'dzeta', 'only_markers', 'particle_on_co', 'circumference', 
#           'orientation', 'R_matrix', 'steps_r_matrix', 'R_matrix_ebe', 'slip_factor', 'momentum_compaction_factor', 'bets0', 'T_rev0', 'gamma0', 
#           'beta0', 'p0c', 'qx', 'qy', 'qs', 'c_minus', 'c_r1_avg', 'c_r2_avg', 'eigenvalues', 'rotation_matrix', 'dmux', 'dmuy', 'bx_chrom', 'by_chrom', 
#           'ax_chrom', 'ay_chrom', 'wx_chrom', 'wy_chrom', 'ddx', 'ddpx', 'ddy', 'ddpy', 'dqx', 'dqy', 'ddqx', 'ddqy', 'eneloss_turn', 'damping_constants_turns', 
#           'damping_constants_s', 'partition_numbers', 'eq_gemitt_x', 'eq_gemitt_y', 'eq_gemitt_zeta', 'eq_nemitt_x', 'eq_nemitt_y', 'eq_nemitt_zeta', 
#           'dl_radiation', 'n_dot_delta_kick_sq_ave', 'values_at', 'k0l', 'k1l', 'k2l', 'k3l', 'k4l', 'k5l', 'k0sl', 'k1sl', 'k2sl', 'k3sl', 'k4sl', 'k5sl', 
#           'angle_rad', 'rot_s_rad', 'hkick', 'vkick', 'ks', 'length', 'element_type', 'isthick', 'parent_name', 'method', 'radiation_method', 'reference_frame', 
#           'line_config', '_action'])


class SynchrotronIntegral:

    def __init__(self, line):

        # The class only needs a line object to be instantiated.
        # The line object contains all the information about the elements, including the Twiss.
        self.line = line
        self.tw = line.twiss(method='4d', strengths=True)#, eneloss_and_damping=True)


        # Properties of the particle -------------------------------------------------------------------------------------------------
        # The values for gamma0 and energy0 are indexed with [0] because they are arrays of shape (1, ).
        # To prevent getting arrays as the output of functions where we just wamt to get scalars, the indexing is performed here.
        self.mass0 = self.line.particle_ref.mass0
        self.gamma0 = self.line.particle_ref.gamma0[0]
        self.energy0 = self.line.particle_ref.energy0[0]
        self.r_e = self.line.particle_ref.get_classical_particle_radius0()

        # Properties of the beam line ------------------------------------------------------------------------------------------------
        self.length = self.tw['length']                 # The length of each element
        self.circum = self.tw['circumference']          # The circumference of the ring
        self.T_0 = self.tw['T_rev0']                    # The revolution period of the beam.

        # Curvatures for each elementwant
        kxy = self._orbit_curvature_()
        self.kx = kxy[0]             # The integrated normal curvature of each element
        self.ky = kxy[1]             # The integrated skew curvature of each element

        # Dispersions ----------------------------------------------------------------------------------------------------------------
        # x-direction
        self.dx   = self.tw['dx']                       # Dispersion in the x-direction
                                                        # Derivative of the dispersion in the x-direction, w.r.t. s
        self.dxprime  = self.tw['dpx'] * (1 - self.tw['delta']) - self.tw['px']

        # y-direction
        self.dy   = self.tw['dy']                       # Dispersion in the y-direction
                                                        # Derivative of the dispersion in the y-direction, w.r.t. s
        self.dyprime  = self.tw['dpy'] * (1 - self.tw['delta']) - self.tw['py']


    # Private methods ----------------------------------------------------------------------------------------------------------------
    # Below are functions that calculate physical quantities that are later used in the integrals.


    # This function calculates the horizontal and vertical acceleration of the particle.
    def _get_derivative_(self, yaxis, xaxis):
            derivative_values = np.zeros(shape=(len(self.length)))

            derivative_values = np.gradient(yaxis, xaxis)

            derivative_values[np.isnan(derivative_values)] = 0

            return derivative_values


    # Calculate the curvature of the design orbit.
    def _orbit_curvature_(self):
        kappa_xy = np.zeros(shape=(2, len(self.length)))
        
        # TODO: I think the self.tw['k0l'] should be replaced with self.tw['angle_rad'].
        angle_rad = self.tw['k0l']
        rot_s_rad = self.tw['rot_s_rad']
        mask = self.length != 0

        kappa_xy[0, :][mask] = angle_rad[mask] * np.cos(rot_s_rad[mask]) / self.length[mask]
        kappa_xy[1, :][mask] = angle_rad[mask] * np.sin(rot_s_rad[mask]) / self.length[mask]

        return kappa_xy


   # Calculate the curvature of the design orbit.
    def _get_curvature_(self):

        angle_rad = self.tw['angle_rad']
        rot_s_rad = self.tw['rot_s_rad']
        x = self.tw['x']
        y = self.tw['y']
        kin_px = self.tw['kin_px']
        kin_py = self.tw['kin_py']
        delta = self.tw['delta']
        length = self.tw['length']

        kappa_0xy = np.zeros(shape=(2, len(self.length)))
        mask = self.length != 0
        kappa_0xy[0, :][mask] = angle_rad[mask] * np.cos(rot_s_rad[mask]) / self.length[mask]
        kappa_0xy[1, :][mask] = angle_rad[mask] * np.sin(rot_s_rad[mask]) / self.length[mask]

        ps = np.sqrt((1 + delta)**2 - kin_px**2 - kin_py**2)
        xp = kin_px / ps
        yp = kin_py / ps

        xp_ele = xp * 0
        yp_ele = yp * 0

        xp_ele[:-1] = (xp[:-1] + xp[1:]) / 2
        yp_ele[:-1] = (yp[:-1] + yp[1:]) / 2

        mask_length = length != 0
        xpp_ele = xp_ele * 0
        ypp_ele = yp_ele * 0
        xpp_ele[mask_length] = np.diff(xp, append=0)[mask_length] / length[mask_length]
        ypp_ele[mask_length] = np.diff(yp, append=0)[mask_length] / length[mask_length]

        k_xy = np.zeros(shape=(2, len(self.length)))

        h = 1 + kappa_0xy[0, :] * x + kappa_0xy[1, :] * y
        hprime = kappa_0xy[0, :] * xp_ele + kappa_0xy[1, :] * yp_ele

        mask1 = xpp_ele**2 + h**2 != 0
        mask2 = xpp_ele**2 + h**2 != 0

        k_xy[0, :] = -(h * (xpp_ele - h * kappa_0xy[0, :]) - 2 * hprime * xp_ele)[mask1] / (xp_ele**2 + h**2)[mask1]**(3/2)
        k_xy[1, :] = -(h * (ypp_ele - h * kappa_0xy[1, :]) - 2 * hprime * yp_ele)[mask2] / (yp_ele**2 + h**2)[mask2]**(3/2)

        return k_xy


    # For the field index
    def _get_fieldindex_(self):
        # The integrated normal and skew quadrupole strengths of each element
        quadkn = self.tw['k1l']            # The integrated normal quadrupole strength of each element
        #quadks = self.tw['k1sl']           # The integrated skew quadrupole strength of each element

        fieldindex = np.zeros(shape=(2, len(self.kx)))
        mask = self.length * self.kx != 0
            
        fieldindex[0, :][mask] = -quadkn[mask] / (self.length * self.kx)[mask]**2
        fieldindex[1, :][mask] = -quadkn[mask] / (self.length * self.kx)[mask]**2
            
        return fieldindex


    # Calculates the H function
    def _H_function_(self):
        betx = self.tw['betx']             # Twiss beta function x
        alfx = self.tw['alfx']             # Twiss alpha x
        gamx = self.tw['gamx']             # Twiss gamma x
        bety = self.tw['bety']             # Twiss beta function y
        alfy = self.tw['alfy']             # Twiss alpha y
        gamy = self.tw['gamy']             # Twiss gamma y

        H_values_xy = np.zeros(shape=(2, len(self.length)))
            
        H_values_xy[0, :] = gamx * self.dx**2 + 2*alfx * self.dx * self.dxprime + betx * self.dxprime**2
        H_values_xy[1, :] = gamy * self.dy**2 + 2*alfy * self.dy * self.dyprime + bety * self.dyprime**2

        return H_values_xy


    # The integrands -----------------------------------------------------------------------------------------------------------------
    # Below, we calculate the integrands. The integrals are arrays of length equal to the number of elements.
    # To get the integrals, we need to sum over al the elements in te integrals.
    # This is done in the methods that calculate the physical quantities.


    # Integrand 1
    # Units: [m]
    # Returns an array of shape (2, number of elements)
    # The two rows correspond to the x- and y-values respectively.
    def _integrand1_(self):
        I1xy_values = np.zeros(shape=(2, len(self.length)))

        I1xy_values[0, :] = self.kx * self.length * self.dx
        I1xy_values[1, :] = self.ky * self.length * self.dy

        return I1xy_values
        
    
    # Integrand 2
    # Units: [m⁻¹]
    def _integrand2_(self):
        I2xy_values = np.zeros(shape=(2, len(self.length)))

        I2xy_values[0, :] = self.length * self.kx**2
        I2xy_values[1, :] = self.length * self.ky**2
        
        return I2xy_values


    # Integrand 3
    # Units: [m⁻²]
    def _integrand3_(self):
        I3xy_values = np.zeros(shape=(2, len(self.length)))

        I3xy_values[0, :] = self.length * np.abs(self.kx)**3
        I3xy_values[1, :] = self.length * np.abs(self.ky)**3

        return I3xy_values
            
    
    # Integrand 4
    # Units: [m⁻¹]
    # Returns an array of shape (2, number of elements)
    # The two rows correspond to the x- and y-values respectively.
    def _integrand4_(self):
        I4xy_values = np.zeros(shape=(2, len(self.length)))

        fieldindex = self._get_fieldindex_()
        
        I4xy_values[0, :] = self.length * self.kx**3 * self.dx * (1 - 2 * fieldindex[0])
        I4xy_values[1, :] = self.length * self.ky**3 * self.dy * (1 - 2 * fieldindex[1])

        return I4xy_values
        
    
    # Integrand 5
    # Units: [m⁻¹]
    def _integrand5_(self):
        I5xy_values = np.zeros(shape=(2, len(self.length)))
        H = self._H_function_()
            
        I5xy_values[0, :] = self.length * np.abs(self.kx)**3 * H[0, :]
        I5xy_values[1, :] = self.length * np.abs(self.ky)**3 * H[1, :]
            
        return I5xy_values


    # The physical quantities --------------------------------------------------------------------------------------------------------
    # Below are the public methods that calculate the physical quantities using the integrands.

    # Momentum Compaction Factor
    # This function gets the (2, number of elements) array of the _Integrand1_
    # and returns a (2, 1) array with the x- and y-values for the momentum compaction factor respectively.
    # Call momentum_compaction()[0] for the x-value
    # and momentum_compaction()[1] for the y-value.
    def momentum_compaction(self):
        I1 = np.sum(self._integrand1_())

        return I1 / self.circum


    # Energy Loss
    # Calculates the energy loss using _Integrand2_().
    # Integrand 2 only depends on the length and bending radii of the elements.
    # This means that for the total energy loss, we can simply sum the x- and y-integrals.
    # This amounts to summing all elements of _Integrand2_() together.
    def energy_loss(self):
        I2 = np.sum(self._integrand2_())

        return 2/3 * self.r_e * self.gamma0**3 * self.energy0 * I2


    # Radiation Damping per Second
    # Calculates the radiation damping using _Integrand2_() and _Integrand4_().
    # I2xy = I2x + I2y, which is just the sum over all elements of _Integrand2_().
    # I4x and I4y are the x- and y-values of the sum over all elements of _Integrand4_().
    # I4xy = I4x + I4y.Equation
    #
    # The radiation damping coefficients are calculated from symmetry considerations.
    # For example, if only horizontal bending etc. is present, alpha_x is proportional to I_2 - I_4.
    # Also, alpha_y is proportional to just I_2.
    # So far, I_2 and I_4 have only been in the x-direction.
    # If we want to include the effects of bending in the y-direction, we can switch the roles of alpha_x and alpha_y.
    # This would result in alpha_x being proportional to I_2 and alpha_y being proportional to I_2 - I_4, which are now calculated in the y-direction.
    # The total effect is assumed to be the sum of the effects in the x- and y-directions.
    # This results in adding an I_2y proportionality to alpha_x and an I_2x proportionality to alpha_y.
    def radiation_damping_s(self):
        rad_damp_coefs = np.zeros(shape=(3,))
        
        I2xy = np.sum(self._integrand2_())
        I4x  = np.sum(self._integrand4_()[0])
        I4y  = np.sum(self._integrand4_()[1])
        I4xy = np.sum(self._integrand4_())
            
        rad_damp_coefs[0] = self.r_e/3 * self.gamma0**3 * clight/self.circum * (I2xy - I4x)
        rad_damp_coefs[1] = self.r_e/3 * self.gamma0**3 * clight/self.circum * (I2xy - I4y)
        rad_damp_coefs[2] = self.r_e/3 * self.gamma0**3 * clight/self.circum * (2 * I2xy + I4xy)

        return rad_damp_coefs
        

    # Radiation Damping per Turn
    # The radiation damping per turn is calculated by multiplying the radiation damping per second with the revolution period.
    def radiation_damping_turns(self):
        return self.radiation_damping_s() * self.T_0


    # RMS Energies
    # It is assumed that the RMS of the energies depends both on the x- and y-integrals.
    def rms_energies(self):
        I2 = np.sum(self._integrand2_())
        I3 = np.sum(self._integrand3_())
        I4 = np.sum(self._integrand4_())

        if I2 - I4 != 0:
            return 55/(32 * 3**(1/2)) * hbar / electron_volt * clight / self.mass0 * self.gamma0**2 * I3 / (2 * I2 - I4)    
        
        else:
            raise ValueError("The denominator of the RMS energy, 2I2 - I4 = 0. The RMS energy is not defined.")
        

    # Radial Quantum Emission
    # Since the denominator depends on the damping contants, it is expected that it contains the same integrals as their respective damping constants.
    # That is, the x-component has I2xy - I4x in the denominator and the y-component has I2xy - I4y.
    def rms_betatron(self):
        rms_betatronxy = np.zeros(shape=(2,))
        
        I2xy = np.sum(self._integrand2_())
        I4x  = np.sum(self._integrand4_(), axis=1)[0]
        I4y  = np.sum(self._integrand4_(), axis=1)[1]
        I5x  = np.sum(self._integrand5_(), axis=1)[0]
        I5y  = np.sum(self._integrand5_(), axis=1)[1]

        if I2xy - I4x != 0:
            rms_betatronxy[0] = 55/(32 * 3**(1/2)) * hbar / electron_volt * clight / self.mass0 * self.gamma0**2 * I5x / (I2xy - I4x)

        if I2xy - I4y != 0:
            rms_betatronxy[1] = 55/(32 * 3**(1/2)) * hbar / electron_volt * clight / self.mass0 * self.gamma0**2 * I5y / (I2xy - I4y)

        return rms_betatronxy
    

    # Equilibrium emittance
    # This quantity is derived from the RMS Energy.
    # See Hoffmann Section 14.3.2 for more information about the equation used here.
    # TODO: Find out how we can correctly implement the effect of phi_s.
        # In the JSON file, phi_s = 180 deg. This value can be used if we replace cos(phi_s) with -cos(phi_s). But is that true in general?
    # TODO: Find out how we can correctly implement the effect of the momentum compaction factor.
        # Do we need to sum both x- and y-values?
    # TODO: Investigate the effect of having multiple cavities.
        # Do we need to simply sum the voltages?
        # What if there are different cavities with different frequencies?
    # TODO: Is creating this large table not a bit too much for such a small function?
    def equilibrium_emittance(self):
        tab = self.line.get_table(attr = True)
        V_cav = tab.rows[tab.element_type == 'Cavity']['voltage']
        f_cav = tab.rows[tab.element_type == 'Cavity']['frequency']

        # self.line.attr['voltage'][mask]
        # self.line.attr['frequency'][mask]

        V_cav_tot = np.sum(V_cav)                           # The voltage of all cavities. See the TODO.
        f_cav = np.sum(f_cav[V_cav != 0])                   # The frequency of the cavity.
        p_0 = self.line.particle_ref.p0c[0] / clight        # The [0] index is to convert this to a scalar instead of a (1, ) array.
        harmonic = f_cav * self.T_0                         # The harmonic number is the ratio of the cavity frequency to the revolution frequency.
        phi_s = 0                                           # The synchronous phase is assumed to be 0. See the TODO.
        alpha_cI = self.momentum_compaction()               # The momentum compaction factor is assumed to be the x-value. See the TODO.

        rms_E_conv_factor = self.T_0 * self.energy0 / p_0 * (alpha_cI * self.energy0 / (2*np.pi * harmonic * V_cav_tot * np.cos(phi_s)))**(1/2)

        return rms_E_conv_factor * self.rms_energies()